import asyncio
import click
from typing import List
from rich.console import Console
from rich.table import Table
from rich.progress import Progress, SpinnerColumn, TextColumn

from targon.client.client import Client
from targon.core.exceptions import TargonError, APIError

console = Console(stderr=True)


def display_error(err: BaseException | str, title: str = "Error") -> None:
    console.print()
    console.print(f"[red]{str(err)}[/red]")
    console.print()


@click.group()
def container():
    """Manage serverless containers."""
    pass


@container.command("ls")
@click.argument("uid", required=False)
@click.option(
    "--sort",
    type=click.Choice(["created", "name", "cost"], case_sensitive=False),
    default="created",
    help="Sort containers by field (default: created)",
)
@click.option(
    "--reverse",
    is_flag=True,
    help="Reverse sort order (descending)",
)
@click.pass_context
def list_containers(
    ctx, uid=None, sort: str = "created", reverse: bool = False
) -> None:
    """List all running containers."""
    client: Client = ctx.obj["client"]

    try:
        with console.status(
            "[bold cyan]Fetching containers...[/bold cyan]", spinner="dots"
        ):
            resources = client.run_async(
                lambda: client.async_serverless.list_container()
            )

        # Display results in a table
        if not resources:
            console.print(
                "\n[dim]No containers found. Deploy one using [cyan]targon deploy[/cyan].[/dim]\n"
            )
            return

        # Sort resources based on the selected field
        def sort_key(resource):
            if sort == "name":
                return (resource.name or "").lower()
            elif sort == "cost":
                return resource.cost if resource.cost is not None else 0
            else:  # created
                return resource.created_at or ""

        resources = sorted(resources, key=sort_key, reverse=reverse)

        table = Table(
            show_header=True,
            header_style="bold cyan",
            border_style="bright_black",
            title="[bold]Serverless Containers[/bold]",
            title_style="bold magenta",
            padding=(0, 1),
        )

        table.add_column("NAME", style="bright_cyan", no_wrap=True)
        table.add_column("UID", style="dim white")
        table.add_column("URL", style="blue")
        table.add_column("COST", justify="right", style="yellow")
        table.add_column("CREATED", style="green")
        total_cost = 0
        for resource in resources:
            total_cost += resource.cost
            table.add_row(
                resource.name or "—",
                resource.uid,
                resource.url if resource.url else "[dim]—[/dim]",
                f"${resource.cost:,.2f}/hr" if resource.cost else "[dim]—[/dim]",
                resource.created_at if resource.created_at else "[dim]—[/dim]",
            )

        console.print()
        console.print(table)
        container_count = len(resources)
        container_text = "container" if container_count == 1 else "containers"
        console.print(
            f"[bold]Total:[/bold] [bright_cyan]{container_count}[/bright_cyan] {container_text} - [yellow]${total_cost:,.2f}/hr[/yellow]"
        )
        console.print()

    except (TargonError, APIError) as e:
        display_error(e, "Failed to list containers")
        raise SystemExit(1)
    except Exception as e:
        display_error(e, "Unexpected error")
        raise SystemExit(1)


@container.command("rm")
@click.argument("container_ids", nargs=-1, required=True)
@click.option("--yes", "-y", is_flag=True, help="Skip confirmation prompt")
@click.pass_context
def remove_containers(ctx, container_ids: List[str], yes: bool) -> None:
    """Delete one or more containers by UID."""
    client: Client = ctx.obj["client"]

    if not yes:
        if len(container_ids) == 1:
            click.confirm(f"Delete container '{container_ids[0]}'?", abort=True)
        else:
            console.print("\n[bold]Containers to delete:[/bold]")
            for cid in container_ids:
                console.print(f"  • [bright_cyan]{cid}[/bright_cyan]")
            console.print()
            click.confirm(f"Delete all {len(container_ids)} containers?", abort=True)

    try:

        async def _delete_all():
            tasks = [
                client.async_serverless.delete_container(cid) for cid in container_ids
            ]
            results = await asyncio.gather(*tasks, return_exceptions=True)
            return {container_ids[i]: results[i] for i in range(len(container_ids))}

        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            console=console,
        ) as progress:
            task_id = progress.add_task(
                f"[cyan]Deleting {len(container_ids)} container(s).[/cyan]",
                total=len(container_ids),
            )
            results = client.run_async(_delete_all)
            progress.update(task_id, completed=len(container_ids))

        console.print()
        failures = []
        for cid, result in results.items():
            if isinstance(result, Exception):
                failures.append(cid)
                console.print(f"[bold red]✖[/bold red] {cid}: {result}")
            else:
                console.print(
                    f"[bold green]✔[/bold green] Deleted [bright_cyan]{cid}[/bright_cyan]"
                )

        console.print()
        if failures:
            console.print(f"[red]Failed:[/red] {len(failures)} / {len(container_ids)}")
            raise SystemExit(1)
        console.print(
            f"[bold green]All {len(container_ids)} containers deleted.[/bold green]\n"
        )

    except (TargonError, APIError) as e:
        display_error(e, "Deletion failed")
        raise SystemExit(1)
    except Exception as e:
        display_error(e, "Unexpected error")
        raise SystemExit(1)



from contextlib import asynccontextmanager
import dataclasses
import asyncio
import sys
import time
import inspect
from pathlib import Path
from typing import AsyncGenerator, Dict, Optional

from targon.client.publish import PublishResponse
from targon.client.app import AppResponse, AppStatusResponse
from targon.client.client import Client
from targon.core.console import Console
from targon.core.objects import BaseApp, _Object
from targon.core.exceptions import ValidationError, ServerlessError
from targon.core.resolver import Resolver
from targon.core.function import _Function
from targon.core.utils import check_object_name
from targon.utils.config_parser import config_to_serverless_requests


@dataclasses.dataclass
class RunningApp:
    app_id: str
    app_name: str
    project_name: str
    function_ids: Dict[str, str] = dataclasses.field(default_factory=dict)
    class_ids: Dict[str, str] = dataclasses.field(default_factory=dict)
    app_page_url: Optional[str] = None


async def _init_local_app_from_name(
    client: Client,
    name: str,
    project_name: str = "default",
) -> RunningApp:

    app_resp: AppResponse = await client.async_app.get_app(
        name=name,
        project_name=project_name,
    )
    return RunningApp(
        app_id=app_resp.AppID,
        app_name=app_resp.Name,
        project_name=app_resp.ProjectName,
        app_page_url=app_resp.WebURL,
    )


async def _create_all_objects(
    client: Client,
    running_app: RunningApp,
    functions: Dict[str, _Function],
    classes: Dict[str, _Function],
    console_instance: Optional[Console] = None,
    app_file: Optional[Path] = None,
    app_module_name: str = "app",
) -> None:

    indexed_objects: Dict[str, _Object] = {**functions, **classes}
    resolver = Resolver(
        client,
        app_id=running_app.app_id,
        project_name=running_app.project_name,
        console_instance=console_instance,
        app_file=app_file,
        app_module_name=app_module_name,
    )

    if console_instance:
        console_instance.step("Creating Objects")

    async def _load(tag: str, obj: _Object) -> None:
        await resolver.load(obj)
        if _Function._is_id_type(obj.object_id):
            running_app.function_ids[tag] = obj.object_id
            if console_instance:
                console_instance.resource(tag, obj.object_id)
        else:
            raise ServerlessError(
                f"Failed to create object: unexpected object ID format",
                function_id=obj.object_id,
            )

    await asyncio.gather(*(_load(tag, obj) for tag, obj in indexed_objects.items()))

    if console_instance:
        function_count = len(running_app.function_ids)
        console_instance.success(
            f"Created {function_count} function{'s' if function_count != 1 else ''}"
        )


async def _check_function_deployment_status(
    client: Client,
    running_app: RunningApp,
    console_instance: Optional[Console] = None,
) -> bool:
    timeout = 300
    poll_interval = 5
    start_time = time.time()

    if console_instance:
        console_instance.substep(f"Awaiting deployment (timeout: {timeout}s)")

    while True:
        elapsed_time = time.time() - start_time

        status_response: AppStatusResponse = await client.async_app.get_app_status(
            app_id=running_app.app_id
        )

        non_deployed = [
            (func_name, func_status.status)
            for func_name, func_status in status_response.functions.items()
            if func_status.status != "deployed"
        ]

        if not non_deployed:
            return True

        if elapsed_time >= timeout:
            if console_instance:
                console_instance.error(f"Deployment timeout after {timeout}s")
                for func_name, status in non_deployed:
                    console_instance.info(f"{func_name}: {status}")
            return False

        if console_instance:
            remaining_time = timeout - elapsed_time
            console_instance.substep(
                f"Checking again in {poll_interval}s ({remaining_time:.0f}s remaining)"
            )

        await asyncio.sleep(poll_interval)


async def _publish_app(
    client: Client,
    running_app: RunningApp,
) -> PublishResponse:
    function_uids = list(running_app.function_ids.values())
    if not function_uids:
        raise ValidationError(
            "No functions were created; nothing to publish.",
            field="functions",
            value=[],
        )

    response = await client.async_publish.publish_serverless(
        app_id=running_app.app_id,
        app_name=running_app.app_name,
        functions=function_uids,
    )

    running_app.app_page_url = response.web_url
    return response


async def deploy_app(
    app: BaseApp,
    client: Client,
    console_instance: Optional[Console] = None,
    name: Optional[str] = None,
    app_file_path: Optional[str] = None,
    running_app: Optional[RunningApp] = None,
) -> None:

    resolved_name = name or app.name or ""
    if not resolved_name or not resolved_name.strip():
        raise ValidationError(
            "You must provide a deployment name. Example: app = targon.App(name=\"my-app\")",
            field="name",
            value=resolved_name,
        )

    check_object_name(resolved_name)

    t0 = time.time()

    if app_file_path:
        # Use the file path provided by the CLI (the actual user script)
        app_file: Optional[Path] = Path(app_file_path).resolve()
        app_module_name = app_file.stem
    else:
        try:
            app_file = Path(inspect.getfile(app.__class__)).resolve()
            app_module_name = app_file.stem
        except (TypeError, OSError):
            app_file = None
            app_module_name = "app"

    if not running_app:
        if console_instance:
            console_instance.step("Initializing app")
        running_app = await _init_local_app_from_name(
            client, resolved_name, project_name="default"
        )
        if console_instance:
            console_instance.success("Initialized")

    await _create_all_objects(
        client,
        running_app,
        app._functions,
        app._classes,
        console_instance=console_instance,
        app_file=app_file,
        app_module_name=app_module_name,
    )
    if console_instance:
        console_instance.step("Publishing app")
    publish_response = await _publish_app(client, running_app)
    if console_instance:
        console_instance.success("Published")

    total_time = time.time() - t0

    if publish_response.summary and console_instance:
        summary_msg = f"Functions: {publish_response.summary.deployed}/{publish_response.summary.total_functions} deployed"
        if publish_response.summary.failed > 0:
            summary_msg += f", {publish_response.summary.failed} failed"

        details_list = [
            f"URL: {running_app.app_page_url}" if running_app.app_page_url else None,
            summary_msg,
            "",
        ]

        # Add invocation instructions for each function
        for tag, func_obj in app._functions.items():
            func_id = running_app.function_ids.get(tag)
            if not func_id:
                continue

            # Check if function has webhook config (is a web endpoint)
            if func_obj._webhook_config:
                details_list.append(f"  • {tag}: {func_obj.web_url}")
            else:
                details_list.append(
                    f"  • {tag}: use 'targon run' with local entry point"
                )

        details_list.extend(
            [
                "",
                "Next steps:",
                f"  • View app state: targon app get {running_app.app_id}",
                f"  • View func logs: targon logs {func_id}",
            ]
        )

        console_instance.final(
            f"Deployed successfully in {total_time:.1f}s!", details_list
        )


@asynccontextmanager
async def run_app(
    app: BaseApp,
    *,
    client: Optional[Client] = None,
    console_instance: Optional[Console] = None,
    name: Optional[str] = None,
    app_file_path: Optional[str] = None,
) -> AsyncGenerator[BaseApp, None]:
    if not app_file_path:
        raise ValidationError(
            "Unable to locate the script file",
            field="app_file_path",
            value=app_file_path,
        )

    if not client:
        client = Client.from_env()

    if console_instance:
        console_instance.step("Initializing app")
    running_app = await _init_local_app_from_name(
        client, name or app.name or "", "default"
    )
    if console_instance:
        console_instance.success("Initialized")

    await deploy_app(app, client, console_instance, name, app_file_path, running_app)

    if console_instance:
        console_instance.step("Checking deployment status")

    all_deployed = await _check_function_deployment_status(
        client, running_app, console_instance
    )

    if all_deployed:
        try:
            if console_instance:
                console_instance.success("All functions deployed")
                console_instance.step("Executing the function")
            yield app

        except Exception as e:
            if console_instance:
                console_instance.error(f"Failed to run app: {e}")
            raise

        finally:
            if console_instance:
                console_instance.success("Local execution completed")
    else:
        if console_instance:
            console_instance.info(
                f"View status: targon app status {running_app.app_id}"
            )
        sys.exit(1)


async def deploy_config(
    config,  # DeployConfig type
    client: Client,
    console_instance: Optional[Console] = None,
) -> RunningApp:
    t0 = time.time()

    # Initialize app
    if console_instance:
        console_instance.step("Initializing app")

    running_app = RunningApp(
        app_id=None,
        app_name=config.app_name,
        project_name="default",
        app_page_url=None,
    )

    if console_instance:
        console_instance.success(f"Parsing app: {config.app_name}")

    # Convert config to serverless requests
    requests = config_to_serverless_requests(config, app_id=running_app.app_id)

    if console_instance:
        console_instance.step(f"Deploying {len(requests)} container(s)")

    # Deploy each container
    deployed_resources = []
    for request in requests:
        try:
            if console_instance:
                console_instance.substep(f"Deploying container: {request.name}")

            resource = await client.async_serverless.deploy_container(request)
            deployed_resources.append(resource)

            if console_instance:
                console_instance.resource(request.name, resource.uid)

        except Exception as e:
            if console_instance:
                console_instance.error(f"Failed to deploy {request.name}: {e}")
            raise

    if console_instance:
        console_instance.success(f"Deployed {len(deployed_resources)} container(s)")

    if console_instance:
        details_list = [
            f"App: {running_app.app_name}",
        ]

        if running_app.app_id:
            details_list.append(f"App ID: {running_app.app_id}")

        details_list.extend(
            [
                "",
                "Deployed containers:",
            ]
        )

        for idx, resource in enumerate(deployed_resources):
            container_name = (
                requests[idx].name if idx < len(requests) else f"container-{idx}"
            )
            details_list.append(
                f"  • {container_name}: https://{resource.uid}.serverless.targon.com"
            )

        details_list.append("")

        if running_app.app_id:
            details_list.extend(
                [
                    "Next steps:",
                    f"  • List containers: targon container ls",
                    "",
                ]
            )

        console_instance.final(f"Deployment completed", details_list)

    return running_app



from dataclasses import asdict, dataclass, field
from typing import Any, Dict, List, Optional

from targon.core.objects import AsyncBaseHTTPClient
from targon.core.exceptions import ValidationError, HydrationError
from targon.client.constants import DEFAULT_BASE_URL, PUBLISH_ENDPOINT


@dataclass
class PublishRequest:
    """Request payload for publishing an app."""

    app_id: str
    app_name: str
    functions: List[str]


@dataclass
class DeploymentSummary:
    total_functions: int
    deployed: int
    failed: int

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "DeploymentSummary":
        if not isinstance(data, dict):
            raise HydrationError(
                f"Invalid data type for DeploymentSummary: expected dict, got {type(data).__name__}",
                object_type="DeploymentSummary",
            )
        return cls(
            total_functions=data.get("total_functions", 0),
            deployed=data.get("deployed", 0),
            failed=data.get("failed", 0),
        )


@dataclass
class FunctionDeploymentStatus:
    id: str
    name: str
    status: str
    invoke_url: Optional[str] = None
    error: Optional[str] = None

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "FunctionDeploymentStatus":
        if not isinstance(data, dict):
            raise HydrationError(
                f"Invalid data type for FunctionDeploymentStatus: expected dict, got {type(data).__name__}",
                object_type="FunctionDeploymentStatus",
            )
        return cls(
            id=data.get("id", ""),
            name=data.get("name", ""),
            status=data.get("status", "unknown"),
            invoke_url=data.get("invoke_url"),
            error=data.get("error"),
        )


@dataclass
class PublishResponse:
    app_id: str
    name: str
    status: str
    web_url: Optional[str]
    summary: Optional[DeploymentSummary]
    functions: List[FunctionDeploymentStatus] = field(default_factory=list)

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "PublishResponse":
        if not isinstance(data, dict):
            raise HydrationError(
                f"Invalid data type for PublishResponse: expected dict, got {type(data).__name__}",
                object_type="PublishResponse",
            )

        summary = None
        if "summary" in data and data["summary"]:
            summary = DeploymentSummary.from_dict(data["summary"])

        functions = []
        if "functions" in data and data["functions"]:
            if not isinstance(data["functions"], list):
                raise HydrationError(
                    "Invalid 'functions' field: expected list",
                    object_type="PublishResponse",
                )
            functions = [
                FunctionDeploymentStatus.from_dict(func) for func in data["functions"]
            ]

        return cls(
            app_id=data.get("app_id", ""),
            name=data.get("name", ""),
            status=data.get("status", "unknown"),
            web_url=data.get("web_url"),
            summary=summary,
            functions=functions,
        )


class AsyncPublishClient(AsyncBaseHTTPClient):
    """Async client for publishing apps to serverless infrastructure."""

    def __init__(self, client):
        super().__init__(client)
        self.base_url = DEFAULT_BASE_URL

    async def publish_serverless(
        self, app_id: str, app_name: str, functions: List[str]
    ) -> PublishResponse:
        if not app_id or not app_id.strip():
            raise ValidationError("app_id is required", field="app_id", value=app_id)

        if not app_name or not app_name.strip():
            raise ValidationError(
                "app_name is required", field="app_name", value=app_name
            )

        if not functions:
            raise ValidationError(
                "functions list cannot be empty", field="functions", value=functions
            )

        if not isinstance(functions, list):
            raise ValidationError(
                "functions must be a list of function UIDs",
                field="functions",
                value=type(functions).__name__,
            )

        for i, func_uid in enumerate(functions):
            if not isinstance(func_uid, str) or not func_uid.strip():
                raise ValidationError(
                    f"Invalid function UID at index {i}: must be a non-empty string",
                    field=f"functions[{i}]",
                    value=func_uid,
                )

        payload = PublishRequest(
            app_id=app_id.strip(),
            app_name=app_name.strip(),
            functions=functions,
        )

        result = await self._async_post(PUBLISH_ENDPOINT, json=asdict(payload))

        if not isinstance(result, dict):
            raise HydrationError(
                f"Invalid response format: expected dict, got {type(result).__name__}",
                object_type="PublishResponse",
            )

        return PublishResponse.from_dict(result)



import click
from targon.cli.auth import get_stored_key, save_api_key
from ..core.console import _rich_console


@click.command("setup", help="Configure Targon CLI with your API key")
def setup():
    _rich_console.print("\n[bold bright_cyan]Targon CLI Setup[/bold bright_cyan]\n")

    # Check if already configured
    existing_key = get_stored_key()
    if existing_key:
        _rich_console.print("[bright_blue]ℹ[/bright_blue] API key already configured.")
        if not click.confirm("Do you want to update it?"):
            _rich_console.print("[dim]Setup cancelled.[/dim]\n")
            return

    # Get API key
    api_key = click.prompt(
        "Enter your Targon API key (starts with 'sn4_')", hide_input=True
    )

    # Validate format
    if not api_key.startswith("sn4_") or len(api_key) != 32:
        _rich_console.print(
            "\n[bold red]✖[/bold red] [bold]Invalid API key format.[/bold] Key should start with 'sn4_' and be 32 characters long.\n"
        )
        return

    if save_api_key(api_key):
        _rich_console.print(
            "\n[bold green]✔[/bold green] [bold]API key saved successfully![/bold]"
        )
        _rich_console.print(
            "[dim italic]  You're ready to use Targon. Try 'targon app list' to get started.[/dim italic]\n"
        )
    else:
        _rich_console.print(
            "\n[bold red]✖[/bold red] [bold]Failed to save API key.[/bold] You can set TARGON_API_KEY environment variable instead.\n"
        )



import os
from typing import Dict, Optional
from targon.client.constants import DEFAULT_BASE_URL
from targon.core.exceptions import ConfigurationError
from targon.version import __version__


class Config:
    api_key: str
    base_url: str
    timeout: int
    max_retries: int
    verify_ssl: bool
    version: str
    user_agent: str
    headers: Dict[str, str]

    def __init__(
        self,
        api_key: Optional[str] = None,
        base_url: str = DEFAULT_BASE_URL,
        timeout: int = 30,
        max_retries: int = 3,
        verify_ssl: bool = True,
        user_agent: Optional[str] = None,
    ) -> None:

        resolved_key = api_key or os.getenv("TARGON_API_KEY")
        if (
            not resolved_key
            or not isinstance(resolved_key, str)
            or not resolved_key.strip()
        ):
            raise ConfigurationError(
                "API key is required. Provide it via the 'api_key' parameter or do targon setup or do "
                "set the TARGON_API_KEY environment variable.",
                config_key="api_key",
            )

        if not base_url or not isinstance(base_url, str) or not base_url.strip():
            raise ConfigurationError(
                "base_url must be a non-empty string", config_key="base_url"
            )

        if timeout <= 0:
            raise ConfigurationError(
                "timeout must be greater than 0", config_key="timeout"
            )

        if max_retries < 0:
            raise ConfigurationError(
                "max_retries must be non-negative", config_key="max_retries"
            )

        self.api_key = resolved_key.strip()
        self.base_url = base_url.rstrip("/")
        self.timeout = timeout
        self.max_retries = max_retries
        self.verify_ssl = verify_ssl
        self.version = __version__
        self.user_agent = (
            user_agent.strip() if user_agent else f"targon-sdk-python/{__version__}"
        )

        self.headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
            "Accept": "application/json",
        }



import aiohttp
from typing import AsyncGenerator, Dict
from targon.core.exceptions import ValidationError, APIError
from targon.client.constants import LOGS_ENDPOINT, DEFAULT_BASE_URL
from targon.core.objects import AsyncBaseHTTPClient
import re


class AsyncLogsClient(AsyncBaseHTTPClient):
    """Async client for streaming function logs."""

    COLORS = [
        '\033[36m',  # Cyan
        '\033[35m',  # Magenta
        '\033[33m',  # Yellow
        '\033[32m',  # Green
        '\033[34m',  # Blue
        '\033[96m',  # Bright Cyan
        '\033[95m',  # Bright Magenta
        '\033[93m',  # Bright Yellow
    ]
    RESET = '\033[0m'
    BOLD = '\033[1m'
    DIM = '\033[2m'

    def __init__(self, client):
        super().__init__(client)
        self.base_url = DEFAULT_BASE_URL
        self._pod_colors: Dict[str, str] = {}
        self._color_index = 0

    def _get_pod_color(self, pod_name: str) -> str:
        """Get or assign a color for a pod."""
        if pod_name not in self._pod_colors:
            self._pod_colors[pod_name] = self.COLORS[
                self._color_index % len(self.COLORS)
            ]
            self._color_index += 1
        return self._pod_colors[pod_name]

    def _extract_pod_name(self, event_line: str) -> str:
        """Extract pod name from event line"""
        match = re.search(r'event:\s*log:\s*(.+)', event_line)
        if match:
            return match.group(1).strip()
        return "unknown"

    def _format_log_line(self, pod_name: str, log_data: str) -> str:
        """Format a log line with pod name and color coding."""
        color = self._get_pod_color(pod_name)

        # Shorten pod name for display (show last 12 chars which is typically the unique part)
        short_pod_name = pod_name[-12:] if len(pod_name) > 12 else pod_name

        # Format: [POD-ID] log message (only pod name is colored)
        formatted = f"{color}{self.BOLD}[{short_pod_name}]{self.RESET} {log_data}"
        return formatted

    async def stream_logs(
        self, serverless_uid: str, follow: bool = True
    ) -> AsyncGenerator[str, None]:
        """Stream logs from a serverless function."""

        if not serverless_uid or not serverless_uid.strip():
            raise ValidationError(
                "serverless_uid is required",
                field="serverless_uid",
                value=serverless_uid,
            )

        payload = {"serverless_uid": serverless_uid.strip(), "follow": follow}

        url = f"{self.base_url}{LOGS_ENDPOINT}"
        timeout = aiohttp.ClientTimeout(
            total=None,  # No total timeout for streaming
            connect=10,  # 10 seconds to establish connection
            sock_read=None,  # No timeout on reading (for long-running streams)
        )

        current_pod = None

        try:
            async with self.session.post(
                url, json=payload, timeout=timeout
            ) as response:
                if response.status >= 400:
                    error_text = await response.text()
                    raise APIError(response.status, error_text)

                # Stream and parse SSE format
                async for line in response.content:
                    if line:
                        decoded_line = line.decode('utf-8').rstrip('\n\r')
                        if not decoded_line:
                            continue

                        if decoded_line.startswith('event:'):
                            current_pod = self._extract_pod_name(decoded_line)

                        elif decoded_line.startswith('data:'):
                            log_data = decoded_line[5:].strip()

                            if log_data and current_pod:
                                # Yield formatted log line
                                yield self._format_log_line(current_pod, log_data)

                        elif not decoded_line.startswith(
                            'event:'
                        ) and not decoded_line.startswith('data:'):
                            if current_pod:
                                yield self._format_log_line(current_pod, decoded_line)
                            else:
                                yield decoded_line

        except aiohttp.ClientError as e:
            raise APIError(
                500, f"Failed to connect to logs endpoint: {str(e)}", cause=e
            )



from typing import Any, Optional, Callable, Dict, List, AsyncIterator
from contextlib import asynccontextmanager

from targon.core.exceptions import ValidationError
from targon.core.resources import Compute
from targon.core.objects import BaseApp, _LocalEntrypoint
from targon.core.partial_function import _PartialFunction
from targon.core.function import _Function
from targon.core.image import _Image


class _App(BaseApp):
    _name: Optional[str]
    _description: Optional[str]
    _tags: Dict[str, str]
    _functions: Dict[str, _Function]
    _classes: Dict[str, Any]
    _image: _Image
    _local_entrypoints: Dict[str, _LocalEntrypoint]
    _web_endpoints: List[str]
    _app_id: Optional[str]
    _project_name: str

    def __init__(self, name: str, image: _Image, project_name: str = "default") -> None:
        if not name or not isinstance(name, str) or not name.strip():
            raise ValidationError(
                "name must be a non-empty string", field="name", value=name
            )

        if not isinstance(image, _Image):
            raise ValidationError(
                "image must be a valid _Image instance",
                field="image",
                value=type(image).__name__,
            )

        if (
            not project_name
            or not isinstance(project_name, str)
            or not project_name.strip()
        ):
            raise ValidationError(
                "project_name must be a non-empty string",
                field="project_name",
                value=project_name,
            )

        self._name = name.strip()
        self._image = image
        self._project_name = project_name.strip()
        self._functions = {}
        self._classes = {}
        self._local_entrypoints = {}
        self._web_endpoints = []
        self._app_id = None

    def function(
        self,
        _warn_parentheses_missing: Any = None,
        *,
        image: Optional[_Image] = None,
        resource: str = Compute.CPU_SMALL,
        min_replicas: int = 1,
        max_replicas: int = 3,
        initial_replicas: int = 0,
        max_concurrency: int | None = None,
        target_concurrency: int | None = None,
        scale_up_delay: str | None = None,
        scale_down_delay: str | None = None,
        zero_grace_period: str | None = None,
        scaling_metric: str | None = None,
        target_value: float | None = None,
        timeout: int = 1500,
        startup_timeout: int = 1500,
        **kwargs: Any,
    ) -> Callable[[Any], _Function]:

        if _warn_parentheses_missing is not None:
            raise ValidationError(
                "Did you forget parentheses? Use `@app.function()`.",
                field="decorator_usage",
            )

        if min_replicas < 0:
            raise ValidationError(
                "min_replicas must be non-negative",
                field="min_replicas",
                value=min_replicas,
            )

        if max_replicas < 1:
            raise ValidationError(
                "max_replicas must be at least 1",
                field="max_replicas",
                value=max_replicas,
            )

        if min_replicas > max_replicas:
            raise ValidationError(
                f"min_replicas ({min_replicas}) cannot exceed max_replicas ({max_replicas})",
                field="min_replicas",
            )

        if timeout < 0:
            raise ValidationError(
                "timeout must be non-negative", field="timeout", value=timeout
            )

        if not resource or not isinstance(resource, str):
            raise ValidationError(
                "resource must be a non-empty string", field="resource", value=resource
            )

        def wrapper(f: Any) -> _Function:
            effective_container_concurrency = max_concurrency
            effective_target_concurrency = target_concurrency

            if isinstance(f, _PartialFunction):
                raw_func = f.raw_f
                webhook_config = f.webhook_config
                if f.is_web_endpoint and raw_func:
                    self._web_endpoints.append(raw_func.__name__)

                # If concurrency has been configured via @targon.concurrent, propagate it
                # into the underlying function's autoscaler configuration.
                if f._params.max_concurrent_inputs is not None:
                    if max_concurrency is not None or target_concurrency is not None:
                        raise ValidationError(
                            "Container concurrency cannot be configured both via "
                            "`@targon.concurrent` and `@app.function(..., max_concurrency=..., "
                            "target_concurrency=...)`.",
                            field="max_concurrency",
                        )
                    effective_container_concurrency = f._params.max_concurrent_inputs
                    effective_target_concurrency = f._params.target_concurrent_inputs
            else:
                raw_func = f
                webhook_config = None

            if raw_func is None:
                raise ValidationError(
                    "Function is None - invalid PartialFunction state", field="function"
                )

            if not callable(raw_func):
                raise ValidationError(
                    "Decorated object must be callable",
                    field="function",
                    value=type(raw_func).__name__,
                )

            tag = getattr(raw_func, "__qualname__", raw_func.__name__)

            fn_obj = _Function.from_local(
                func=raw_func,
                app=self,
                image=image or self._image,
                name=tag,
                webhook_config=webhook_config,
                user_cls=None,
                resource_name=resource,
                min_replicas=min_replicas,
                max_replicas=max_replicas,
                initial_replicas=initial_replicas,
                max_concurrency=effective_container_concurrency,
                target_concurrency=effective_target_concurrency,
                scale_up_delay=scale_up_delay,
                scale_down_delay=scale_down_delay,
                zero_grace_period=zero_grace_period,
                scaling_metric=scaling_metric,
                target_value=target_value,
                timeout=timeout,
                startup_timeout=startup_timeout,
                **kwargs,
            )

            self._functions[tag] = fn_obj

            return fn_obj

        return wrapper

    def local_entrypoint(
        self, _warn_parentheses_missing: Any = None, *, name: Optional[str] = None
    ) -> Callable[[Callable[..., Any]], _LocalEntrypoint]:
        if _warn_parentheses_missing is not None:
            raise ValidationError(
                "Did you forget parentheses? Suggestion: `@app.local_entrypoint()`.",
                field="decorator_usage",
            )

        if name is not None:
            if not isinstance(name, str):
                raise ValidationError(
                    "name must be a string", field="name", value=type(name).__name__
                )
            if not name.strip():
                raise ValidationError(
                    "name cannot be an empty string", field="name", value=name
                )

        def wrapped(raw_f: Callable[..., Any]) -> _LocalEntrypoint:
            if not callable(raw_f):
                raise ValidationError(
                    "Decorated object must be callable",
                    field="function",
                    value=type(raw_f).__name__,
                )

            tag = name.strip() if name is not None else raw_f.__qualname__

            if tag in self._local_entrypoints:
                raise ValidationError(
                    f"Duplicate local entrypoint name: {tag}. Local entrypoint names must be unique.",
                    field="name",
                    value=tag,
                )

            entrypoint = self._local_entrypoints[tag] = _LocalEntrypoint(raw_f, self)
            return entrypoint

        return wrapped

    def deploy(self) -> "_App":
        return self

    def hydrate(self, app_id: str) -> None:
        if not app_id or not isinstance(app_id, str) or not app_id.strip():
            raise ValidationError(
                "app_id must be a non-empty string", field="app_id", value=app_id
            )
        self._app_id = app_id.strip()

    @asynccontextmanager
    async def run(self) -> AsyncIterator["_App"]:
        import inspect
        from targon.core.executor import run_app
        from targon.core.console import Console

        frame = inspect.currentframe()
        if frame and frame.f_back and frame.f_back.f_back:
            caller_frame = frame.f_back.f_back
            app_file_path = caller_frame.f_globals.get('__file__')
        else:
            app_file_path = None

        c = Console(self.name)
        with c:
            async with run_app(
                app=self, console_instance=c, app_file_path=app_file_path
            ):
                yield self


App = _App



from ..core.objects import BaseApp
from ..core.app import App
from ..core.exceptions import TargonError
from ..core.console import console

import importlib.util
import sys
import dataclasses
import inspect
from typing import Optional
from pathlib import Path
import click
from rich.markdown import Markdown


@dataclasses.dataclass
class ImportRef:
    file_path: str
    object_path: Optional[str] = None


# This is the object that will
# be searched if nothing is mentioned
DEFAULT_APP_NAME = "app"


def parse_import_ref(file_ref: str) -> ImportRef:
    """
    Parse a file reference like 'examples/getting_started.py' or 'examples/getting_started.py::main'
    """
    if '::' in file_ref:
        file_path, object_path = file_ref.split('::', 1)
    else:
        file_path, object_path = file_ref, ""
    return ImportRef(file_path=file_path, object_path=object_path)


# Hey Python, I have got this random file somewhere on my computer.
# Please load it as a module so I can use its functions/classes.
def import_file_or_module(import_ref: ImportRef, base_cmd: str = ""):
    if "" not in sys.path:
        sys.path.insert(0, "")

    full_path = Path(import_ref.file_path).resolve()
    if "." in full_path.name.removesuffix(".py"):
        raise ValueError(
            f"Invalid source filename: {full_path.name!r}."
            "\n\nSource filename cannot contain additional period characters."
        )
    sys.path.insert(0, str(full_path.parent))

    module_name = inspect.getmodulename(import_ref.file_path)
    assert module_name is not None
    spec = importlib.util.spec_from_file_location(module_name, import_ref.file_path)
    assert spec is not None
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    try:
        assert spec.loader
        spec.loader.exec_module(module)
    except Exception as exc:
        raise TargonError(f"Failed to execute module {full_path}: {exc}") from exc

    return module


def import_app_from_ref(
    import_ref: ImportRef, base_cmd: str = "targon deploy"
) -> BaseApp:
    import_path = import_ref.file_path
    object_path = import_ref.object_path or DEFAULT_APP_NAME

    module = import_file_or_module(import_ref, base_cmd)
    if "." in object_path:
        raise click.UsageError(
            f"Nested object paths are not supported. Use simple object names like '{object_path.split('.')[0]}'"
        )
    # from module import object_path
    app = getattr(module, object_path)

    if app is None:
        console.print(
            f"[bold red]Could not find Targon app '{object_path}' in {import_path}.[/bold red]"
        )

        if not object_path:
            guidance_msg = Markdown(
                f"Expected to find an app variable named **`{DEFAULT_APP_NAME}`** (the default app name). "
                "If your `App` is assigned to a different variable name, "
                "you must specify it in the app ref argument. "
                f"For example an App variable `app_2 = App()` in `{import_path}` would "
                f"be specified as `{import_path}::app_2`."
            )
            console.print(guidance_msg)

        sys.exit(1)

    if not isinstance(app, App):
        raise click.UsageError(f"{app} is not a Targon App")

    return app



import click
import sys
import asyncio
from typing import Union
from rich.console import Console
from targon import Client
from targon.cli.auth import get_api_key
from targon.core.exceptions import TargonError, APIError
from targon.cli.inventory import capacity
from targon.cli.setup import setup
from targon.cli.deploy import deploy
from targon.cli.run import run
from targon.cli.app import app
from targon.cli.logs import logs
from targon.cli.container import container
from targon.version import __version__

console = Console(stderr=True)


def display_error(err: Union[Exception, str], title: str = "Error"):
    """Display a formatted error message."""
    console.print()
    console.print(f"[red]{str(err)}[/red]")
    console.print()


def _cleanup_client(client: Client):
    """Cleanup function to close client sessions on exit."""
    def cleanup():
        client.close()
        if client._async_session is not None and not client._async_session.closed:
            try:
                asyncio.run(client._async_session.close())
            except Exception:
                pass  
    return cleanup


class SafeGroup(click.Group):
    """A Click Group that catches Targon errors and displays them nicely."""

    def __call__(self, *args, **kwargs):
        try:
            return super().__call__(*args, **kwargs)
        except (TargonError, APIError) as e:
            display_error(e, "Error")
            sys.exit(1)
        except Exception as e:
            if isinstance(e, SystemExit):
                raise
            # Let Click handle Abort (Ctrl+C)
            if isinstance(e, click.Abort):
                raise

            display_error(e, "Unexpected Error")
            sys.exit(1)


@click.group(cls=SafeGroup, invoke_without_command=True)
@click.version_option(__version__, prog_name="Targon CLI")
@click.pass_context
def cli(ctx):
    """Targon SDK CLI - Interact with Targon for secure compute."""
    ctx.ensure_object(dict)
    
    # If no subcommand is provided, print help
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())
        return
    
    # Skip client initialization for commands that don't need auth
    if ctx.invoked_subcommand in ("setup", "completion"):
        return
    
    try:
        resolved_api_key = get_api_key()
        client = Client(api_key=resolved_api_key)
        ctx.obj["client"] = client
        ctx.call_on_close(_cleanup_client(client))
    except Exception as e:
        # Catch auth errors specifically if they happen during init
        if isinstance(e, (TargonError, APIError)):
            raise
        # For other errors during setup, we might want to let them bubble
        # to the SafeGroup handler, but providing context is good.
        raise TargonError(f"Failed to initialize client: {e}") from e


@cli.command()
@click.option(
    "-s", "--shell",
    type=click.Choice(["bash", "zsh", "fish"]),
    help="Shell type (auto-detected if not provided).",
)
def completion(shell: str):
    """Install shell completion to virtualenv."""
    import os
    from pathlib import Path
    
    # Auto-detect shell if not provided
    if not shell:
        shell = Path(os.environ.get("SHELL", "bash")).name
        if shell not in ("bash", "zsh", "fish"):
            shell = "bash"
    
    venv = os.environ.get("VIRTUAL_ENV")
    if not venv:
        raise click.ClickException("Not in a virtualenv. Activate one first.")
    
    activate = Path(venv) / "bin" / "activate"
    if not activate.exists():
        raise click.ClickException(f"Activate script not found: {activate}")
    
    content = activate.read_text()
    
    if "_TARGON_COMPLETE" in content:
        click.echo("Completions already installed.")
        return
    
    line = f'\neval "$(_TARGON_COMPLETE={shell}_source targon)"\n'
    activate.write_text(content + line)
    click.echo(f"Installed completions to {activate}. Run: source {activate}")


# Register commands
cli.add_command(setup)
cli.add_command(capacity)
cli.add_command(deploy, name="deploy")
cli.add_command(run, name="run")
cli.add_command(app, name="app")
cli.add_command(logs, name="logs")
cli.add_command(container, name="container")

if __name__ == '__main__':
    cli()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: function_execution.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC, 6, 31, 1, '', 'function_execution.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x18\x66unction_execution.proto\x12\x0etargon.runtime\"\xf3\x01\n\x0e\x45xecuteRequest\x12\x12\n\nrequest_id\x18\x01 \x01(\t\x12\x0c\n\x04\x61rgs\x18\x02 \x01(\x0c\x12\x0e\n\x06kwargs\x18\x03 \x01(\x0c\x12\x10\n\x08input_id\x18\x04 \x01(\t\x12\x18\n\x10\x66unction_call_id\x18\x05 \x01(\t\x12\x12\n\ntimeout_ms\x18\x06 \x01(\x03\x12>\n\x08metadata\x18\x07 \x03(\x0b\x32,.targon.runtime.ExecuteRequest.MetadataEntry\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xec\x02\n\x0f\x45xecuteResponse\x12\x12\n\nrequest_id\x18\x01 \x01(\t\x12\x36\n\x06status\x18\x02 \x01(\x0e\x32&.targon.runtime.ExecuteResponse.Status\x12\x0e\n\x06result\x18\x03 \x01(\x0c\x12\r\n\x05\x65rror\x18\x04 \x01(\t\x12\x11\n\ttraceback\x18\x05 \x01(\t\x12\x19\n\x11\x65xecution_time_ms\x18\x06 \x01(\x03\x12?\n\x08metadata\x18\x07 \x03(\x0b\x32-.targon.runtime.ExecuteResponse.MetadataEntry\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"N\n\x06Status\x12\x0b\n\x07SUCCESS\x10\x00\x12\t\n\x05\x45RROR\x10\x01\x12\x0b\n\x07TIMEOUT\x10\x02\x12\r\n\tCANCELLED\x10\x03\x12\x10\n\x0cRATE_LIMITED\x10\x04\"%\n\x12HealthCheckRequest\x12\x0f\n\x07service\x18\x01 \x01(\t\"\xa9\x01\n\x13HealthCheckResponse\x12\x41\n\x06status\x18\x01 \x01(\x0e\x32\x31.targon.runtime.HealthCheckResponse.ServingStatus\"O\n\rServingStatus\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0b\n\x07SERVING\x10\x01\x12\x0f\n\x0bNOT_SERVING\x10\x02\x12\x13\n\x0fSERVICE_UNKNOWN\x10\x03\x32^\n\x10\x46unctionExecutor\x12J\n\x07\x45xecute\x12\x1e.targon.runtime.ExecuteRequest\x1a\x1f.targon.runtime.ExecuteResponse2\xae\x01\n\x06Health\x12P\n\x05\x43heck\x12\".targon.runtime.HealthCheckRequest\x1a#.targon.runtime.HealthCheckResponse\x12R\n\x05Watch\x12\".targon.runtime.HealthCheckRequest\x1a#.targon.runtime.HealthCheckResponse0\x01\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'function_execution_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals['_EXECUTEREQUEST_METADATAENTRY']._loaded_options = None
    _globals['_EXECUTEREQUEST_METADATAENTRY']._serialized_options = b'8\001'
    _globals['_EXECUTERESPONSE_METADATAENTRY']._loaded_options = None
    _globals['_EXECUTERESPONSE_METADATAENTRY']._serialized_options = b'8\001'
    _globals['_EXECUTEREQUEST']._serialized_start = 45
    _globals['_EXECUTEREQUEST']._serialized_end = 288
    _globals['_EXECUTEREQUEST_METADATAENTRY']._serialized_start = 241
    _globals['_EXECUTEREQUEST_METADATAENTRY']._serialized_end = 288
    _globals['_EXECUTERESPONSE']._serialized_start = 291
    _globals['_EXECUTERESPONSE']._serialized_end = 655
    _globals['_EXECUTERESPONSE_METADATAENTRY']._serialized_start = 241
    _globals['_EXECUTERESPONSE_METADATAENTRY']._serialized_end = 288
    _globals['_EXECUTERESPONSE_STATUS']._serialized_start = 577
    _globals['_EXECUTERESPONSE_STATUS']._serialized_end = 655
    _globals['_HEALTHCHECKREQUEST']._serialized_start = 657
    _globals['_HEALTHCHECKREQUEST']._serialized_end = 694
    _globals['_HEALTHCHECKRESPONSE']._serialized_start = 697
    _globals['_HEALTHCHECKRESPONSE']._serialized_end = 866
    _globals['_HEALTHCHECKRESPONSE_SERVINGSTATUS']._serialized_start = 787
    _globals['_HEALTHCHECKRESPONSE_SERVINGSTATUS']._serialized_end = 866
    _globals['_FUNCTIONEXECUTOR']._serialized_start = 868
    _globals['_FUNCTIONEXECUTOR']._serialized_end = 962
    _globals['_HEALTH']._serialized_start = 965
    _globals['_HEALTH']._serialized_end = 1139
# @@protoc_insertion_point(module_scope)



import grpc.aio
import cloudpickle
import uuid
import time
import logging
import asyncio
from dataclasses import dataclass
from typing import Any, Optional, Dict, Tuple
from enum import Enum
from collections import OrderedDict

from targon.proto import function_execution_pb2
from targon.proto import function_execution_pb2_grpc
from targon.core.exceptions import TargonError

logger = logging.getLogger(__name__)


class FunctionInvocationError(TargonError):
    """Raised when function invocation fails."""

    def __init__(self, message: str, traceback: Optional[str] = None):
        super().__init__(message)
        self.traceback = traceback


class FunctionTimeoutError(TargonError):
    """Raised when function execution times out."""

    pass


class CircuitBreakerState(Enum):
    """Circuit breaker states."""

    CLOSED = "closed"
    OPEN = "open"
    HALF_OPEN = "half_open"


class CircuitBreaker:
    """Circuit breaker pattern implementation for fault tolerance."""

    def __init__(
        self,
        failure_threshold: int = 5,
        recovery_timeout: float = 30.0,
        success_threshold: int = 2,
    ) -> None:
        self.failure_threshold = failure_threshold
        self.recovery_timeout = recovery_timeout
        self.success_threshold = success_threshold

        self.state = CircuitBreakerState.CLOSED
        self.failure_count = 0
        self.success_count = 0
        self.last_failure_time: Optional[float] = None
        self._lock = asyncio.Lock()

    async def call(self, func, *args, **kwargs) -> Any:
        """Execute function with circuit breaker protection."""
        async with self._lock:
            if self.state == CircuitBreakerState.OPEN:
                if (
                    self.last_failure_time
                    and (time.time() - self.last_failure_time) >= self.recovery_timeout
                ):
                    logger.info(
                        "Circuit breaker transitioning to HALF_OPEN (testing recovery)"
                    )
                    self.state = CircuitBreakerState.HALF_OPEN
                    self.success_count = 0
                else:
                    raise FunctionInvocationError(
                        f"Circuit breaker is OPEN (too many failures). "
                        f"Retry after {self.recovery_timeout}s recovery period."
                    )

        try:
            result = await func(*args, **kwargs)
            await self._on_success()
            return result
        except Exception as e:
            await self._on_failure()
            raise

    async def _on_success(self) -> None:
        """Handle successful execution."""
        async with self._lock:
            self.failure_count = 0

            if self.state == CircuitBreakerState.HALF_OPEN:
                self.success_count += 1
                if self.success_count >= self.success_threshold:
                    logger.info("Circuit breaker transitioning to CLOSED (recovered)")
                    self.state = CircuitBreakerState.CLOSED
                    self.success_count = 0

    async def _on_failure(self) -> None:
        """Handle failed execution."""
        async with self._lock:
            self.failure_count += 1
            self.last_failure_time = time.time()

            if self.state == CircuitBreakerState.HALF_OPEN:
                logger.warning(
                    "Circuit breaker transitioning to OPEN (test request failed)"
                )
                self.state = CircuitBreakerState.OPEN
                self.success_count = 0
            elif self.state == CircuitBreakerState.CLOSED:
                if self.failure_count >= self.failure_threshold:
                    logger.warning(
                        f"Circuit breaker transitioning to OPEN "
                        f"({self.failure_count} consecutive failures)"
                    )
                    self.state = CircuitBreakerState.OPEN


class GRPCConnectionPool:
    """Connection pool for gRPC channels with LRU eviction."""

    def __init__(self, max_channels: int = 100) -> None:
        self.max_channels = max_channels
        self._channels: OrderedDict[Tuple[str, int], grpc.aio.Channel] = OrderedDict()
        self._stubs: Dict[
            Tuple[str, int], function_execution_pb2_grpc.FunctionExecutorStub
        ] = {}
        self._circuit_breakers: Dict[Tuple[str, int], CircuitBreaker] = {}
        self._lock = asyncio.Lock()

    async def get_stub(
        self,
        host: str,
        port: int,
        credentials: Optional[grpc.ChannelCredentials] = None,
    ) -> Tuple[function_execution_pb2_grpc.FunctionExecutorStub, CircuitBreaker]:
        """Get or create a secure gRPC stub for the given endpoint."""
        key = (host, port)

        async with self._lock:
            if key in self._channels:
                self._channels.move_to_end(key)
                return self._stubs[key], self._circuit_breakers[key]

            address = f"{host}:{port}"
            options = [
                ('grpc.max_send_message_length', 100 * 1024 * 1024),
                ('grpc.max_receive_message_length', 100 * 1024 * 1024),
                ('grpc.keepalive_time_ms', 30000),
                ('grpc.keepalive_timeout_ms', 10000),
                ('grpc.http2.max_pings_without_data', 0),
                ('grpc.keepalive_permit_without_calls', 1),
            ]

            creds = credentials or grpc.ssl_channel_credentials()
            channel = grpc.aio.secure_channel(address, creds, options=options)
            logger.debug(f"Created secure channel to {address}")

            stub = function_execution_pb2_grpc.FunctionExecutorStub(channel)
            circuit_breaker = CircuitBreaker()

            self._channels[key] = channel
            self._stubs[key] = stub
            self._circuit_breakers[key] = circuit_breaker

            if len(self._channels) > self.max_channels:
                oldest_key = next(iter(self._channels))
                old_channel = self._channels.pop(oldest_key)
                self._stubs.pop(oldest_key)
                self._circuit_breakers.pop(oldest_key)
                await old_channel.close()
                logger.debug(f"Evicted channel from pool: {oldest_key}")

            logger.debug(f"Created new channel in pool: {key}")
            return stub, circuit_breaker

    async def close_all(self) -> None:
        """Close all channels in the pool."""
        async with self._lock:
            for channel in self._channels.values():
                await channel.close()
            self._channels.clear()
            self._stubs.clear()
            self._circuit_breakers.clear()


_connection_pool = GRPCConnectionPool()


class GRPCFunctionClient:
    """Async gRPC client for invoking functions over secure TLS/SSL connections."""

    def __init__(
        self,
        host: str,
        port: int = 50051,
        timeout: int = 300,
        use_connection_pool: bool = True,
        credentials: Optional[grpc.ChannelCredentials] = None,
    ) -> None:
        self.host = host
        self.port = port
        self.timeout = timeout
        self.use_connection_pool = use_connection_pool
        self.credentials = credentials

    async def invoke(
        self,
        *args,
        timeout: Optional[int] = None,
        request_id: Optional[str] = None,
        **kwargs,
    ) -> Any:
        """Invoke the function asynchronously with the given arguments."""
        if request_id is None:
            request_id = str(uuid.uuid4())

        timeout_seconds = timeout or self.timeout
        timeout_ms = int(timeout_seconds * 1000)

        try:
            args_bytes = cloudpickle.dumps(args) if args else b""
            kwargs_bytes = cloudpickle.dumps(kwargs) if kwargs else b""
        except Exception as e:
            logger.error(f"[{request_id}] Failed to serialize arguments: {e}")
            raise FunctionInvocationError(f"Failed to serialize arguments: {e}")

        request = function_execution_pb2.ExecuteRequest(
            request_id=request_id,
            args=args_bytes,
            kwargs=kwargs_bytes,
            timeout_ms=timeout_ms,
        )

        if self.use_connection_pool:
            stub, circuit_breaker = await _connection_pool.get_stub(
                self.host, self.port, self.credentials
            )
        else:
            stub, circuit_breaker = await self._create_direct_stub()

        try:
            result = await circuit_breaker.call(
                self._execute_request, stub, request, request_id, timeout_seconds
            )
            return result
        except (FunctionInvocationError, FunctionTimeoutError):
            raise
        except Exception as e:
            logger.error(
                f"[{request_id}] Unexpected error during invocation: {e}", exc_info=True
            )
            raise FunctionInvocationError(f"Unexpected error: {e}")

    async def _execute_request(
        self,
        stub: function_execution_pb2_grpc.FunctionExecutorStub,
        request: function_execution_pb2.ExecuteRequest,
        request_id: str,
        timeout_seconds: float,
    ) -> Any:
        """Execute the gRPC request."""
        start_time = time.time()

        try:
            logger.debug(
                f"[{request_id}] Invoking function via gRPC at {self.host}:{self.port}"
            )
            response = await stub.Execute(request, timeout=timeout_seconds + 5)

        except grpc.RpcError as e:
            elapsed = time.time() - start_time
            logger.error(
                f"[{request_id}] gRPC error after {elapsed:.2f}s: "
                f"code={e.code()}, details={e.details()}"
            )

            if e.code() == grpc.StatusCode.DEADLINE_EXCEEDED:
                raise FunctionTimeoutError(
                    f"Function call timed out after {elapsed:.2f}s (timeout: {timeout_seconds}s)"
                )
            raise FunctionInvocationError(f"gRPC error: {e.details()}")

        elapsed_ms = int((time.time() - start_time) * 1000)
        logger.debug(f"[{request_id}] Response received in {elapsed_ms}ms")

        if response.status == function_execution_pb2.ExecuteResponse.SUCCESS:
            try:
                result = cloudpickle.loads(response.result)
                logger.info(
                    f"[{request_id}] ✅ Success "
                    f"(execution: {response.execution_time_ms}ms, total: {elapsed_ms}ms)"
                )
                return result
            except Exception as e:
                logger.error(f"[{request_id}] Failed to deserialize result: {e}")
                raise FunctionInvocationError(f"Failed to deserialize result: {e}")

        elif response.status == function_execution_pb2.ExecuteResponse.TIMEOUT:
            logger.error(f"[{request_id}] Function execution timeout: {response.error}")
            raise FunctionTimeoutError(response.error)

        elif response.status == function_execution_pb2.ExecuteResponse.ERROR:
            error_msg = response.error
            if response.traceback:
                error_msg += f"\n\nRemote traceback:\n{response.traceback}"
            logger.error(f"[{request_id}] Function execution error: {error_msg}")
            raise FunctionInvocationError(error_msg, traceback=response.traceback)

        else:
            logger.error(f"[{request_id}] Unknown response status: {response.status}")
            raise FunctionInvocationError(f"Unknown status: {response.status}")

    async def _create_direct_stub(
        self,
    ) -> Tuple[function_execution_pb2_grpc.FunctionExecutorStub, CircuitBreaker]:
        """Create a direct stub without using the connection pool."""
        address = f"{self.host}:{self.port}"
        options = [
            ('grpc.max_send_message_length', 100 * 1024 * 1024),
            ('grpc.max_receive_message_length', 100 * 1024 * 1024),
        ]

        creds = self.credentials or grpc.ssl_channel_credentials()
        channel = grpc.aio.secure_channel(address, creds, options=options)

        stub = function_execution_pb2_grpc.FunctionExecutorStub(channel)
        circuit_breaker = CircuitBreaker()
        return stub, circuit_breaker

    async def health_check(self) -> bool:
        """Check if the function server is healthy."""
        try:
            stub, _ = await _connection_pool.get_stub(
                self.host, self.port, self.credentials
            )

            request = function_execution_pb2.HealthCheckRequest()

            address = f"{self.host}:{self.port}"
            creds = self.credentials or grpc.ssl_channel_credentials()
            channel = grpc.aio.secure_channel(address, creds)

            health_stub = function_execution_pb2_grpc.HealthStub(channel)

            response = await health_stub.Check(request, timeout=5)
            await channel.close()

            return response.status == function_execution_pb2.HealthCheckResponse.SERVING
        except Exception as e:
            logger.debug(f"Health check failed: {e}")
            return False

    async def __aenter__(self) -> "GRPCFunctionClient":
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        pass



"""
Minimal runtime __init__.py for targon containers.

This module provides stub classes for Image and App that allow user app modules
to import and execute module-level code without errors, even though the actual
functionality isn't needed at runtime (image already built, app already configured).
"""

import contextlib


class _StubImage:
    def __init__(self, *args, **kwargs):
        pass

    def __getattr__(self, name):
        """Return a method that returns self for any attribute access."""

        def method(*args, **kwargs):
            return self

        return method

    @contextlib.contextmanager
    def imports(self):
        yield

    @classmethod
    def debian_slim(cls, python_version=None):
        """Stub debian_slim class method."""
        return cls()

    @classmethod
    def from_registry(cls, *args, **kwargs):
        """Stub from_registry class method."""
        return cls()

    @classmethod
    def from_dockerfile(cls, *args, **kwargs):
        """Stub from_dockerfile class method."""
        return cls()

    def pip_install(self, *args, **kwargs):
        """Stub pip_install method."""
        return self

    def apt_install(self, *args, **kwargs):
        """Stub apt_install method."""
        return self

    def pip_install_from_requirements(self, *args, **kwargs):
        """Stub pip_install_from_requirements method."""
        return self

    def add_local_file(self, *args, **kwargs):
        """Stub add_local_file method."""
        return self

    def add_local_dir(self, *args, **kwargs):
        """Stub add_local_dir method."""
        return self

    def env(self, *args, **kwargs):
        """Stub env method."""
        return self

    def workdir(self, *args, **kwargs):
        """Stub workdir method."""
        return self

    def with_runtime(self, *args, **kwargs):
        """Stub with_runtime method."""
        return self

    def run_commands(self, *args, **kwargs):
        """Stub run_commands method."""
        return self


class _StubApp:
    def __init__(self, *args, **kwargs):
        """Accept any arguments to match the real App signature."""
        pass

    def function(self, *args, **kwargs):
        def decorator(func):
            return func

        return decorator

    def local_entrypoint(self, *args, **kwargs):
        def decorator(func):
            return func

        return decorator


class _StubCompute:
    """Stub Compute class with tier name constants."""

    # CPU Tiers
    CPU_SMALL = "cpu-small"
    CPU_MEDIUM = "cpu-medium"
    CPU_LARGE = "cpu-large"
    CPU_XL = "cpu-xl"

    # GPU Tiers
    # H200s
    H200_SMALL = "h200-small"
    H200_MEDIUM = "h200-medium"
    H200_LARGE = "h200-large"
    H200_XL = "h200-xl"
    # H100s
    H100_SMALL = "h100-small"
    H100_MEDIUM = "h100-medium"
    H100_LARGE = "h100-large"
    H100_XL = "h100-xl"
    # B200s
    B200_SMALL = "b200-small"
    B200_MEDIUM = "b200-medium"
    B200_LARGE = "b200-large"
    B200_XL = "b200-xl"
    # RTX4090s
    RTX4090_SMALL="rtx4090-small"
    RTX4090_MEDIUM="rtx4090-medium"
    RTX4090_LARGE="rtx4090-large"
    
def _stub_decorator(*args, **kwargs):
    def decorator(func):
        return func

    # Support both @decorator and @decorator()
    if len(args) == 1 and callable(args[0]) and not kwargs:
        return args[0]
    return decorator


fastapi_endpoint = _stub_decorator
asgi_app = _stub_decorator
wsgi_app = _stub_decorator
web_server = _stub_decorator
concurrent = _stub_decorator
enter = _stub_decorator
exit = _stub_decorator
method = _stub_decorator

Image = _StubImage
App = _StubApp
Compute = _StubCompute

__version__ = "0.1.0"
__all__ = [
    "Image",
    "App",
    "Compute",
    "fastapi_endpoint",
    "asgi_app",
    "wsgi_app",
    "web_server",
    "enter",
    "exit",
    "method",
    "concurrent",
    "__version__",
]



import asyncio
import os
import tempfile
from asyncio import Future
from collections.abc import Hashable
from typing import TYPE_CHECKING, Optional
from pathlib import Path

from targon.core.exceptions import ValidationError

if TYPE_CHECKING:
    from targon.core.objects import _Object
    from targon.client.client import Client


class Resolver:
    """Resolver for loading and hydrating Targon objects."""

    _local_uuid_to_future: dict[str, Future["_Object"]]
    _project_name: Optional[str]
    _app_id: Optional[str]
    _deduplication_cache: dict[Hashable, Future["_Object"]]
    _client: "Client"
    _build_start: float
    _console_instance: Optional[object]
    _app_file: Optional[Path]
    _app_module_name: Optional[str]

    def __init__(
        self,
        client: "Client",
        *,
        project_name: Optional[str] = None,
        app_id: Optional[str] = None,
        console_instance: Optional[object] = None,
        app_file: Optional[Path] = None,
        app_module_name: Optional[str] = None,
    ):
        self._local_uuid_to_future = {}
        self._client = client
        self._app_id = app_id
        self._project_name = project_name
        self._deduplication_cache = {}
        self._console_instance = console_instance
        self._app_file = app_file
        self._app_module_name = app_module_name

        with tempfile.TemporaryFile() as temp_file:
            # Use file mtime to track build start time
            self._build_start = os.fstat(temp_file.fileno()).st_mtime

    @property
    def app_id(self) -> Optional[str]:
        return self._app_id

    @property
    def client(self) -> "Client":
        return self._client

    @property
    def project_name(self) -> Optional[str]:
        return self._project_name

    @property
    def build_start(self) -> float:
        return self._build_start

    @property
    def console_instance(self) -> Optional[object]:
        return self._console_instance

    async def load(
        self, obj: "_Object", existing_object_id: Optional[str] = None
    ) -> "_Object":
        """Load and hydrate an object, handling dependencies and deduplication."""
        if getattr(obj, "_is_hydrated", False):
            if obj.local_uuid not in self._local_uuid_to_future:
                fut: Future["_Object"] = Future()
                fut.set_result(obj)
                self._local_uuid_to_future[obj.local_uuid] = fut
            return obj

        # Handle deduplication
        deduplication_key: Optional[Hashable] = None
        if obj._deduplication_key:
            deduplication_key = await obj._deduplication_key()

        cached_future = self._local_uuid_to_future.get(obj.local_uuid)

        # Check deduplication cache
        # check if the object is already in the deduplication cache
        # which means that this object has been loaded before other app
        if not cached_future and deduplication_key is not None:
            cached_future = self._deduplication_cache.get(deduplication_key)
            if cached_future:
                hydrated_object = await cached_future
                obj._hydrate(
                    hydrated_object.object_id,
                    self._client,
                    hydrated_object._get_metadata(),
                )
                return obj

        if not cached_future:

            async def loader() -> "_Object":
                deps = obj.deps()
                if deps:
                    await asyncio.gather(*[self.load(dep) for dep in deps])

                if not obj._load:
                    raise ValidationError(
                        f"Object {obj} has no loader function",
                        field="obj._load",
                    )

                await obj._load(obj, self, existing_object_id)

                if (
                    existing_object_id is not None
                    and existing_object_id.startswith("fu-")
                    and obj.object_id != existing_object_id
                ):
                    raise ValidationError(
                        f"Object ID mismatch: expected {existing_object_id}, got {obj.object_id}",
                        field="object_id",
                        value=obj.object_id,
                    )

                return obj

            cached_future = asyncio.create_task(loader())
            self._local_uuid_to_future[obj.local_uuid] = cached_future

            if deduplication_key is not None:
                self._deduplication_cache[deduplication_key] = cached_future

        return await cached_future

    def objects(self) -> list["_Object"]:
        """Get all loaded objects."""
        unique_objects: dict[str, "_Object"] = {}
        for fut in self._local_uuid_to_future.values():
            if not fut.done():
                raise ValidationError(
                    "Cannot retrieve objects: not all objects have been resolved yet"
                )
            obj = fut.result()
            unique_objects.setdefault(obj.object_id, obj)
        return list(unique_objects.values())

    def clear_cache(self):
        self._deduplication_cache.clear()

    def get_object_count(self) -> int:
        return len(self._local_uuid_to_future)

    def is_loading(self, obj: "_Object") -> bool:
        return obj.local_uuid in self._local_uuid_to_future

    async def load_all(self, objects: list["_Object"]) -> list["_Object"]:
        return await asyncio.gather(*[self.load(obj) for obj in objects])

    def get_loading_status(self) -> dict[str, str]:
        status: dict[str, str] = {}
        for uuid, future in self._local_uuid_to_future.items():
            if future.done():
                try:
                    obj = future.result()
                    status[uuid] = f"loaded: {obj.object_id}"
                except Exception as e:
                    status[uuid] = f"error: {type(e).__name__}: {e}"
            else:
                status[uuid] = "loading"
        return status

    def get_app_file(self) -> Optional[Path]:
        return self._app_file

    def get_app_module_name(self) -> Optional[str]:
        return self._app_module_name



from dataclasses import dataclass
from typing import Any, Dict, Optional, List, Tuple
import asyncio

from targon.core.objects import AsyncBaseHTTPClient
from targon.core.exceptions import TargonError, ValidationError
from targon.core.partial_function import WebhookConfig
from targon.client.functions import AutoscalerSettings
from targon.client.constants import (
    DEFAULT_BASE_URL,
    GET_APP_ENDPOINT,
    GET_APP_STATUS_ENDPOINT,
    LIST_APPS_ENDPOINT,
    DELETE_APP_ENDPOINT,
    LIST_FUNCTIONS_ENDPOINT,
    GET_FUNCTION_BY_ID_ENDPOINT,
)


@dataclass(slots=True)
class AppRequest:
    Name: str
    ProjectName: str


@dataclass(slots=True)
class AppResponse:
    AppID: str
    Name: str
    ProjectID: str
    ProjectName: str
    Created: bool
    WebURL: str


@dataclass(slots=True)
class FunctionStatus:
    name: str
    uid: str
    url: str
    status: str


@dataclass(slots=True)
class AppStatusResponse:
    app_id: str
    name: str
    project_id: str
    project_name: str
    function_count: int
    functions: Dict[str, FunctionStatus]
    created_at: str
    updated_at: str


@dataclass(slots=True)
class AppListItem:
    app_id: str
    name: str
    project_id: str
    created_at: str
    updated_at: str


@dataclass(slots=True)
class ListAppsResponse:
    apps: List[AppListItem]
    total: int


@dataclass(slots=True)
class FunctionItem:
    uid: str
    name: str
    module: Optional[str]
    qualname: Optional[str]
    image_id: Optional[str]
    created_at: str
    updated_at: str


@dataclass(slots=True)
class ListFunctionsResponse:
    app_id: str
    functions: List[FunctionItem]
    total: int


@dataclass(slots=True)
class FunctionDetailResponse:
    uid: str
    app_id: str
    name: str
    module: Optional[str]
    qualname: Optional[str]
    image_id: Optional[str]
    resource_name: Optional[str]
    webhook_config: Optional[WebhookConfig]
    autoscaler_settings: Optional[AutoscalerSettings]
    timeout_secs: Optional[int]
    startup_timeout: Optional[int]
    url: str
    status: Optional[str]
    created_at: str
    updated_at: str


class AsyncAppClient(AsyncBaseHTTPClient):
    """Async client for app registration and management."""

    def __init__(self, client: Any) -> None:
        super().__init__(client)
        self.base_url = DEFAULT_BASE_URL

    async def get_app(self, name: str, project_name: str = "default") -> AppResponse:
        if not name or not name.strip():
            raise ValidationError("App name cannot be empty", field="name")

        if not project_name or not project_name.strip():
            raise ValidationError("Project name cannot be empty", field="project_name")

        payload: Dict[str, str] = {"name": name, "project_name": project_name}

        result = await self._async_post(GET_APP_ENDPOINT, json=payload)
        if not isinstance(result, dict):
            raise TargonError(f"Unexpected response format: {type(result).__name__}")

        return AppResponse(
            AppID=result.get("app_id", ""),
            Name=result.get("name", name),
            ProjectID=result.get("project_id", ""),
            ProjectName=result.get("project_name", project_name),
            Created=result.get("created", False),
            WebURL=result.get("web_url", ""),
        )

    async def get_app_status(self, app_id: str) -> AppStatusResponse:
        if not app_id or not app_id.strip():
            raise ValidationError("App ID cannot be empty", field="app_id")

        endpoint = GET_APP_STATUS_ENDPOINT.format(app_id=app_id)
        result = await self._async_get(endpoint)
        if not isinstance(result, dict):
            raise TargonError(f"Unexpected response format: {type(result).__name__}")

        functions_data = result.get("functions", {})
        if not isinstance(functions_data, dict):
            raise TargonError(
                f"Expected functions to be dict, got {type(functions_data).__name__}"
            )

        functions = {
            func_id: FunctionStatus(
                name=func.get("name", ""),
                uid=func.get("uid", ""),
                url=func.get("url", ""),
                status=func.get("status", ""),
            )
            for func_id, func in functions_data.items()
        }

        return AppStatusResponse(
            app_id=result.get("app_id", app_id),
            name=result.get("name", ""),
            project_id=result.get("project_id", ""),
            project_name=result.get("project_name", ""),
            function_count=result.get("function_count", 0),
            functions=functions,
            created_at=result.get("created_at", ""),
            updated_at=result.get("updated_at", ""),
        )

    async def list_apps(self) -> ListAppsResponse:
        """List all apps that are currently deployed/running or recently stopped."""
        result = await self._async_get(LIST_APPS_ENDPOINT)

        if not isinstance(result, dict):
            raise TargonError(f"Unexpected response format: {type(result).__name__}")

        apps_data = result.get("apps", [])
        if not isinstance(apps_data, list):
            raise TargonError(
                f"Expected apps to be list, got {type(apps_data).__name__}"
            )

        apps = [
            AppListItem(
                app_id=app.get("app_id", ""),
                name=app.get("name", ""),
                project_id=app.get("project_id", ""),
                created_at=app.get("created_at", ""),
                updated_at=app.get("updated_at", ""),
            )
            for app in apps_data
        ]

        return ListAppsResponse(
            apps=apps,
            total=result.get("total", len(apps)),
        )

    async def delete_app(self, app_id: str) -> Dict[str, Any]:
        """Delete an app and all corresponding deployments."""
        if not app_id or not app_id.strip():
            raise ValidationError("App ID cannot be empty", field="app_id")

        endpoint = DELETE_APP_ENDPOINT.format(app_id=app_id)
        result = await self._async_delete(endpoint)

        if not isinstance(result, dict):
            raise TargonError(f"Unexpected response format: {type(result).__name__}")

        return result

    async def list_functions(self, app_id: str) -> ListFunctionsResponse:
        """List all functions for a given app."""
        if not app_id or not app_id.strip():
            raise ValidationError("App ID cannot be empty", field="app_id")

        endpoint = LIST_FUNCTIONS_ENDPOINT.format(app_id=app_id)
        result = await self._async_get(endpoint)

        if not isinstance(result, dict):
            raise TargonError(f"Unexpected response format: {type(result).__name__}")

        functions_data = result.get("functions", [])
        if not isinstance(functions_data, list):
            raise TargonError(
                f"Expected functions to be list, got {type(functions_data).__name__}"
            )

        functions = [
            FunctionItem(
                uid=func.get("uid", ""),
                name=func.get("name", ""),
                module=func.get("module"),
                qualname=func.get("qualname"),
                image_id=func.get("image_id"),
                created_at=func.get("created_at", ""),
                updated_at=func.get("updated_at", ""),
            )
            for func in functions_data
        ]

        return ListFunctionsResponse(
            app_id=result.get("app_id", app_id),
            functions=functions,
            total=result.get("total", len(functions)),
        )

    async def get_function(self, function_id: str) -> FunctionDetailResponse:
        """Get detailed information about a specific function by its UID."""
        if not function_id or not function_id.strip():
            raise ValidationError("Function ID cannot be empty", field="function_id")

        # Use the direct function lookup endpoint
        endpoint = GET_FUNCTION_BY_ID_ENDPOINT.format(function_id=function_id)
        result = await self._async_get(endpoint)

        if not isinstance(result, dict):
            raise TargonError(f"Unexpected response format: {type(result).__name__}")

        # Parse webhook_config if present
        webhook_config = None
        if webhook_config_data := result.get("webhook_config"):
            webhook_config = WebhookConfig(**webhook_config_data)

        # Parse autoscaler_settings if present
        autoscaler_settings = None
        if autoscaler_data := result.get("autoscaler_settings"):
            autoscaler_settings = AutoscalerSettings(**autoscaler_data)

        return FunctionDetailResponse(
            uid=result.get("uid", function_id),
            app_id=result.get("app_id", ""),
            name=result.get("name", ""),
            module=result.get("module"),
            qualname=result.get("qualname"),
            image_id=result.get("image_id"),
            resource_name=result.get("resource_name"),
            webhook_config=webhook_config,
            autoscaler_settings=autoscaler_settings,
            timeout_secs=result.get("timeout_secs"),
            startup_timeout=result.get("startup_timeout"),
            url=result.get("url", ""),
            status=result.get("status"),
            created_at=result.get("created_at", ""),
            updated_at=result.get("updated_at", ""),
        )

    async def list_apps_with_details(
        self,
    ) -> Tuple[ListAppsResponse, List[Optional[AppStatusResponse]]]:
        apps_response = await self.list_apps()

        if not apps_response.apps:
            return apps_response, []

        # Fetch detailed status for each app in parallel
        tasks = [self.get_app_status(app.app_id) for app in apps_response.apps]
        results = await asyncio.gather(*tasks, return_exceptions=True)

        # Process results
        detailed_apps = []
        for res in results:
            if isinstance(res, Exception):
                detailed_apps.append(None)
            else:
                detailed_apps.append(res)

        return apps_response, detailed_apps



import re
import os
import shlex
import contextlib
import warnings
from dataclasses import dataclass
from typing import Callable, Dict, List, Optional, Sequence, Tuple, Union
from pathlib import Path, PurePosixPath

from targon.core.objects import _Object
from targon.core.exceptions import ValidationError
from targon.core.resolver import Resolver


@dataclass
class DockerfileSpec:
    commands: List[str]
    context_files: Dict[str, str]


def _dockerfile_function_rep(rep: str) -> Callable:
    def decorator(func: Callable) -> Callable:
        func.rep = rep
        return func

    return decorator


def _make_pip_install_args(
    find_links: Optional[str] = None,
    index_url: Optional[str] = None,
    extra_index_url: Optional[str] = None,
    pre: bool = False,
    extra_options: str = "",
) -> str:
    flags = [
        ("--find-links", find_links),
        ("--index-url", index_url),
        ("--extra-index-url", extra_index_url),
    ]

    args = " ".join(
        f"{flag} {shlex.quote(value)}" for flag, value in flags if value is not None
    )
    if pre:
        args += " --pre"

    if extra_options:
        if args:
            args += " "
        args += f"{extra_options}"

    return args


def _flatten_str_args(
    function_name: str, arg_name: str, args: Sequence[Union[str, List[str]]]
) -> List[str]:
    """Takes a sequence of strings, or string lists, and flattens it."""
    ret = []
    for x in args:
        if isinstance(x, str):
            ret.append(x)
        elif isinstance(x, list) and all(isinstance(y, str) for y in x):
            ret.extend(x)
        else:
            raise ValidationError(
                f"{function_name}: {arg_name} must only contain strings"
            )
    return ret


def _create_ignore_function(
    ignore: Union[Sequence[str], Callable[[Path], bool]],
) -> Callable[[Path], bool]:
    """Create an ignore function from various input types"""
    if callable(ignore):
        return ignore

    if not ignore:  # Empty sequence
        return lambda p: False

    # Convert patterns to simple matching
    patterns = list(ignore)

    def pattern_matcher(path: Path) -> bool:
        """Check if path matches any ignore pattern."""
        path_str = str(path)
        name = path.name

        for pattern in patterns:
            # Simple pattern matching
            if pattern.startswith("*"):
                # Suffix match: *.pyc
                if name.endswith(pattern[1:]):
                    return True
            elif pattern.endswith("*"):
                # Prefix match: __pycache__*
                if name.startswith(pattern[:-1]):
                    return True
            elif "*" in pattern:
                # Contains match: *test*
                clean_pattern = pattern.replace("*", "")
                if clean_pattern in name:
                    return True
            else:
                # Exact match
                if name == pattern or path_str.endswith(pattern):
                    return True

        return False

    return pattern_matcher


def _collect_directory_files(
    directory: Path, ignore_fn: Callable[[Path], bool]
) -> List[Tuple[Path, Path]]:
    """Recursively collect all files from a directory.

    Args:
        directory: Root directory to scan
        ignore_fn: Function to check if a path should be ignored

    Returns:
        List of tuples (absolute_path, relative_path)
    """
    collected = []

    for root, dirs, files in os.walk(directory):
        root_path = Path(root)

        # Filter out ignored directories (modifies dirs in-place to prevent descent)
        dirs[:] = [d for d in dirs if not ignore_fn(root_path / d)]

        for file in files:
            file_path = root_path / file

            # Skip ignored files
            if ignore_fn(file_path):
                continue

            # Calculate relative path
            try:
                rel_path = file_path.relative_to(directory)
                collected.append((file_path, rel_path))
            except ValueError:
                continue

    return collected


class _Image(_Object, type_prefix="img"):
    def _initialize_from_empty(self) -> None:
        self._registry_ref = None
        self._collected_base_images = {}
        self._collected_dockerfile_function = None
        self._runtime_injected = False

    def _get_metadata(self) -> Optional[Dict]:
        return {"registry_ref": self._registry_ref}

    def _hydrate_metadata(self, metadata: Optional[Dict]) -> None:
        if metadata is not None:
            self._registry_ref = metadata.get("registry_ref")

    def _get_dockerfile_function(self) -> Optional[Callable]:
        return getattr(self, '_collected_dockerfile_function', None)

    @staticmethod
    def _from_args(
        *,
        base_images: Optional[Dict[str, "_Image"]] = None,
        dockerfile_function: Optional[Callable[[], DockerfileSpec]] = None,
        force_build: bool = False,
    ) -> "_Image":

        # This means this is the base Image
        if base_images is None:
            base_images = {}

        def _deps() -> List[_Image]:
            return []

        def _collect_all_dockerfile_parts(
            current_base_images: Dict[str, "_Image"],
            current_dockerfile_function: Optional[Callable],
        ) -> Tuple[List[str], Dict[str, str]]:
            """Recursively collect all Dockerfile commands and context files from the image chain."""
            all_commands = []
            all_context_files = {}

            for _, base_img in current_base_images.items():
                base_df_func = (
                    base_img._get_dockerfile_function()
                    if hasattr(base_img, '_get_dockerfile_function')
                    else None
                )
                base_base_images = getattr(base_img, '_collected_base_images', {})

                base_commands, base_context = _collect_all_dockerfile_parts(
                    base_base_images, base_df_func
                )
                all_commands.extend(base_commands)
                all_context_files.update(base_context)

            if current_dockerfile_function:
                spec = current_dockerfile_function()

                all_commands.extend(spec.commands)
                all_context_files.update(spec.context_files)

            return all_commands, all_context_files

        async def _load(self, resolver, existing_object_id: Optional[str]):
            """
            This function executes during hydration.
            """

            client = resolver.client
            console_instance = resolver.console_instance

            # Auto-inject runtime if we have app context
            if resolver._app_file and not self._runtime_injected:
                # Create a new image with runtime injected
                runtime_image = self._inject_runtime(
                    resolver._app_file, resolver._app_module_name
                )
                # Use the runtime image's collected parts instead
                all_commands, all_context_files = _collect_all_dockerfile_parts(
                    runtime_image._collected_base_images,
                    runtime_image._collected_dockerfile_function,
                )
            else:
                # Collect all Dockerfile commands and context files from the entire chain
                # base_images and dockerfile_function are already stored on self during _from_args
                all_commands, all_context_files = _collect_all_dockerfile_parts(
                    self._collected_base_images, self._collected_dockerfile_function
                )

            if not all_commands:
                raise ValidationError(
                    "No commands provided for image", field="commands"
                )

            # Read file contents for context files
            # context_files maps filename -> filepath, we need to read the content
            hydrated_context_files = {}
            for filename, filepath in all_context_files.items():
                with open(filepath, 'r') as f:
                    hydrated_context_files[filename] = f.read()

            # Start the build step
            if console_instance:
                console_instance.step("Building image")

            # Use Heim client to build the image with NDJSON streaming
            try:
                registry_ref = await client.async_heim.build_image(
                    app_id=resolver.app_id,
                    dockerfile_commands=all_commands,
                    context_files=hydrated_context_files,
                    console_instance=console_instance,
                )

                # Create a stable local ID from the registry reference
                # Format: "185-209-179-245.nip.io/img-ci1tzd1ba079:20251027213343"
                safe_ref = (
                    registry_ref.split("/")[1].replace("/", "-").replace(":", "-")
                )
                object_id = f"{safe_ref}"

            except Exception as e:
                if console_instance:
                    console_instance.error("Image build failed", str(e))
                raise ValidationError(f"Image build failed: {e}", field="build")

            self._hydrate(object_id, client, {"registry_ref": registry_ref})

        # Create representation for debugging
        rep = f"Image({dockerfile_function.rep if dockerfile_function else None})"

        img = _Image._from_loader(_load, rep, deps=_deps)

        # Store base_images and dockerfile_function so they can be accessed during collection
        # These are set immediately upon creation, not just during _load
        img._collected_base_images = base_images
        img._collected_dockerfile_function = dockerfile_function

        # Ensure _runtime_injected flag is initialized if not already set
        if not hasattr(img, '_runtime_injected'):
            img._runtime_injected = False

        return img

    def apt_install(
        self,
        *packages: Union[str, List[str]],
    ) -> "_Image":
        pkgs = _flatten_str_args("apt_install", "packages", list(packages))
        if not pkgs:
            return self

        package_args = " ".join(pkgs)

        @_dockerfile_function_rep("apt_install")
        def build_dockerfile():
            commands = [
                "RUN apt-get update",
                f"RUN apt-get install -y {package_args}",
            ]
            return DockerfileSpec(commands=commands, context_files={})

        return _Image._from_args(
            base_images={'base': self},
            dockerfile_function=build_dockerfile,
        )

    def pip_install(
        self,
        *packages: Union[str, List[str]],
        find_links: Optional[str] = None,
        index_url: Optional[str] = None,
        extra_index_url: Optional[str] = None,
        pre: bool = False,
        extra_options: str = "",
    ) -> "_Image":

        pkgs = _flatten_str_args("pip_install", "packages", list(packages))
        if not pkgs:
            return self

        if any(pkg.startswith("-") for pkg in pkgs):
            raise ValidationError(
                "Package list for `Image.pip_install` cannot contain other arguments; try the `extra_options` parameter instead.",
                field="packages",
            )

        @_dockerfile_function_rep("pip_install")
        def build_dockerfile():
            package_args = " ".join(pkgs)
            extra_args = _make_pip_install_args(
                find_links, index_url, extra_index_url, pre, extra_options
            )
            commands = [f"RUN python -m pip install {package_args} {extra_args}"]
            return DockerfileSpec(commands=commands, context_files={})

        return _Image._from_args(
            base_images={"base": self},
            dockerfile_function=build_dockerfile,
        )

    def pip_install_from_requirements(
        self,
        requirements_txt: str,
        find_links: Optional[str] = None,
        *,
        index_url: Optional[str] = None,
        extra_index_url: Optional[str] = None,
        pre: bool = False,
        extra_options: str = "",
    ) -> "_Image":
        # Expand user path and validate file exists
        requirements_txt_path = os.path.expanduser(requirements_txt)
        requirements_path = Path(requirements_txt_path).resolve()

        if not requirements_path.exists():
            raise ValidationError(f"Requirements file not found: {requirements_txt}")

        if not requirements_path.is_file():
            raise ValidationError(f"Path is not a file: {requirements_txt}")

        context_filename = ".requirements.txt"

        @_dockerfile_function_rep("pip_install_from_requirements")
        def build_dockerfile():
            context_files = {context_filename: str(requirements_path)}
            extra_args = _make_pip_install_args(
                find_links, index_url, extra_index_url, pre, extra_options
            )

            commands = [
                f"COPY {context_filename} /.requirements.txt",
                f"RUN python -m pip install -r /.requirements.txt {extra_args}".strip(),
            ]

            return DockerfileSpec(commands=commands, context_files=context_files)

        return _Image._from_args(
            base_images={"base": self},
            dockerfile_function=build_dockerfile,
        )

    def dockerfile_commands(
        self,
        *dockerfile_commands: Union[str, List[str]],
        context_files: Dict[str, str] = {},
    ) -> "_Image":
        if not isinstance(context_files, dict):
            raise ValidationError(
                "context_files must be a dictionary",
                field="context_files",
                value=type(context_files).__name__,
            )

        cmds = _flatten_str_args(
            "dockerfile_commands", "dockerfile_commands", list(dockerfile_commands)
        )
        if not cmds:
            return self

        @_dockerfile_function_rep("dockerfile_commands")
        def build_dockerfile():
            return DockerfileSpec(commands=[*cmds], context_files=context_files)

        return _Image._from_args(
            base_images={"base": self},
            dockerfile_function=build_dockerfile,
        )

    def entrypoint(
        self,
        entrypoint_commands: List[str],
    ) -> "_Image":
        if not isinstance(entrypoint_commands, list) or not all(
            isinstance(x, str) for x in entrypoint_commands
        ):
            raise ValidationError("entrypoint_commands must be a list of string")

        args_str = _flatten_str_args(
            "entrypoint", "entrypoint_commands", entrypoint_commands
        )
        args_str = '"' + '", "'.join(args_str) + '"' if args_str else ""

        @_dockerfile_function_rep("entrypoint")
        def build_dockerfile():
            dockerfile_cmd = f"ENTRYPOINT [{args_str}]"
            return DockerfileSpec(commands=[dockerfile_cmd], context_files={})

        return _Image._from_args(
            base_images={"base": self},
            dockerfile_function=build_dockerfile,
        )

    def run_commands(
        self,
        *commands: Union[str, List[str]],
    ) -> "_Image":
        cmds = _flatten_str_args("run_commands", "commands", list(commands))
        if not cmds:
            return self

        @_dockerfile_function_rep("run_commands")
        def build_dockerfile() -> DockerfileSpec:
            return DockerfileSpec(
                commands=[f"RUN {cmd}" for cmd in cmds], context_files={}
            )

        return _Image._from_args(
            base_images={"base": self},
            dockerfile_function=build_dockerfile,
        )

    def env(self, vars: Dict[str, str]) -> "_Image":
        if not isinstance(vars, dict):
            raise ValidationError(
                "env() expects a dictionary", field="vars", value=type(vars).__name__
            )

        if not all(
            isinstance(key, str) and isinstance(value, str)
            for key, value in vars.items()
        ):
            raise ValidationError("ENV keys and values must be strings", field="vars")

        @_dockerfile_function_rep("env")
        def build_dockerfile() -> DockerfileSpec:
            env_commands = [
                f"ENV {key}={shlex.quote(val)}" for (key, val) in vars.items()
            ]
            return DockerfileSpec(commands=env_commands, context_files={})

        return _Image._from_args(
            base_images={"base": self},
            dockerfile_function=build_dockerfile,
        )

    def workdir(self, path: Union[str, PurePosixPath]) -> "_Image":
        @_dockerfile_function_rep("workdir")
        def build_dockerfile() -> DockerfileSpec:
            commands = [f"WORKDIR {shlex.quote(str(path))}"]
            return DockerfileSpec(commands=commands, context_files={})

        return _Image._from_args(
            base_images={"base": self},
            dockerfile_function=build_dockerfile,
        )

    def cmd(self, cmd: List[str]) -> "_Image":
        if not isinstance(cmd, list) or not all(isinstance(x, str) for x in cmd):
            raise ValidationError("Image CMD must be a list of strings", field="cmd")

        cmd_str = _flatten_str_args("cmd", "cmd", cmd)
        cmd_str = '"' + '", "'.join(cmd_str) + '"' if cmd_str else ""

        @_dockerfile_function_rep("cmd")
        def build_dockerfile() -> DockerfileSpec:
            dockerfile_cmd = f"CMD [{cmd_str}]"
            return DockerfileSpec(commands=[dockerfile_cmd], context_files={})

        return _Image._from_args(
            base_images={"base": self},
            dockerfile_function=build_dockerfile,
        )

    @contextlib.contextmanager
    def imports(self):
        is_remote = (
            os.environ.get("TARGON_ENVIRONMENT") == "remote"
            or os.environ.get("APP_MODULE") is not None
        )

        try:
            yield
        except Exception as exc:
            # If we're in remote environment, always raise
            if is_remote:
                raise

            if isinstance(exc, ImportError):
                pass
            else:
                warnings.warn(
                    f"Warning: Caught a non-ImportError exception in an `image.imports()` block: {repr(exc)}\n"
                    f"This might indicate a problem in your import code beyond just missing packages."
                )

    @staticmethod
    def from_dockerfile(
        path: Union[str, Path],
        *,
        context_dir: Optional[Union[Path, str]] = None,
        ignore: Union[Sequence[str], Callable[[Path], bool]] = [],
    ) -> "_Image":

        # Expand and validate Dockerfile path
        dockerfile_path = Path(os.path.expanduser(path)).resolve()

        if not dockerfile_path.exists():
            raise ValidationError(f"Dockerfile not found: {path}")

        if not dockerfile_path.is_file():
            raise ValidationError(f"Path is not a file: {path}")

        # Determine context directory
        if context_dir is None:
            # Default to directory containing the Dockerfile
            resolved_context_dir = dockerfile_path.parent
        else:
            resolved_context_dir = Path(os.path.expanduser(context_dir)).resolve()

            if not resolved_context_dir.exists():
                raise ValidationError(f"Context directory not found: {context_dir}")

            if not resolved_context_dir.is_dir():
                raise ValidationError(f"Context path is not a directory: {context_dir}")

        # Create ignore function
        ignore_fn = _create_ignore_function(ignore)

        # Collect context files
        collected_files = _collect_directory_files(resolved_context_dir, ignore_fn)

        # Create context files mapping
        context_files = {}
        for local_file_path, rel_path in collected_files:
            # Use relative path as-is for context files to match Dockerfile COPY commands
            context_files[str(rel_path)] = str(local_file_path)

        @_dockerfile_function_rep("from_dockerfile")
        def build_dockerfile():
            # Read Dockerfile contents
            with open(os.path.expanduser(path), 'r') as f:
                dockerfile_contents = f.read()

            # Split into commands (lines)
            commands = [line for line in dockerfile_contents.split('\n')]

            return DockerfileSpec(commands=commands, context_files=context_files)

        return _Image._from_args(
            dockerfile_function=build_dockerfile,
        )

    @staticmethod
    def debian_slim(python_version: Optional[str] = None) -> "_Image":
        """Create a Debian-based slim image."""
        if python_version is None:
            python_version = "3.11"

        # Handle float inputs (common mistake)
        if isinstance(python_version, float):
            python_version = str(python_version)

        # Create base image tag
        base_image = f"python:{python_version}-slim"

        @_dockerfile_function_rep("debain_slim")
        def build_dockerfile():
            commands = [
                f"FROM {base_image}",
                "RUN apt-get update",
                "RUN apt-get install -y gcc gfortran build-essential",
                "RUN pip install --upgrade pip",
                "RUN echo 'debconf debconf/frontend select Noninteractive' | debconf-set-selections",
            ]
            return DockerfileSpec(commands=commands, context_files={})

        return _Image._from_args(
            dockerfile_function=build_dockerfile,
        )

    @staticmethod
    def _registry_setup_commands(
        tag: str,
        setup_commands: List[str],
        add_python: Optional[str] = None,
    ) -> List[str]:
        """Generate Dockerfile commands for registry images with optional Python installation"""
        if not add_python:
            return [f"FROM {tag}", *setup_commands]

        # Validate Python version format (major.minor only)
        if not re.match(r'^3\.\d{1,2}$', add_python):
            raise ValidationError(
                f"Invalid Python version: {add_python}. "
                "Expected format like '3.11', '3.12', etc."
            )

        major, minor = add_python.split('.')
        minor_int = int(minor)

        # Build package list - distutils is deprecated in 3.12+
        python_packages = [
            f"python{add_python}",
            f"python{add_python}-dev",
            f"python{add_python}-venv",
        ]

        # Only add distutils for Python < 3.12
        if minor_int < 12:
            python_packages.append(f"python{add_python}-distutils")

        packages_str = " ".join(python_packages)

        commands = [
            f"FROM {tag}",
            # Run any user-provided setup commands first
            *setup_commands,
            # Install Python using deadsnakes PPA (for Ubuntu/Debian)
            # Use DEBIAN_FRONTEND=noninteractive to avoid prompts
            "ENV DEBIAN_FRONTEND=noninteractive",
            "RUN apt-get update",
            "RUN apt-get install -y software-properties-common",
            "RUN add-apt-repository -y ppa:deadsnakes/ppa",
            "RUN apt-get update",
            f"RUN apt-get install -y {packages_str}",
            # Install pip using get-pip.py (most reliable method)
            "RUN apt-get install -y curl ca-certificates",
            "RUN curl -sS https://bootstrap.pypa.io/get-pip.py -o /tmp/get-pip.py",
            f"RUN python{add_python} /tmp/get-pip.py",
            "RUN rm /tmp/get-pip.py",
            # Create symlinks so 'python' and 'pip' work without version suffix
            f"RUN ln -sf /usr/bin/python{add_python} /usr/local/bin/python",
            f"RUN ln -sf /usr/bin/python{add_python} /usr/local/bin/python3",
            f"RUN ln -sf /usr/local/bin/pip{major} /usr/local/bin/pip || true",
            # Set environment
            "ENV PATH=/usr/local/bin:$PATH",
            # Verify installation
            "RUN python --version && pip --version",
            # Clean up
            "RUN apt-get clean && rm -rf /var/lib/apt/lists/*",
        ]

        return commands

    @staticmethod
    def from_registry(
        tag: str,
        *,
        setup_dockerfile_commands: List[str] = [],
        force_build: bool = False,
        add_python: Optional[str] = None,
    ) -> "_Image":
        """Build a Targon Image from a public or private image registry"""
        if not tag or not isinstance(tag, str) or not tag.strip():
            raise ValidationError(
                "tag must be a non-empty string", field="tag", value=tag
            )

        if not isinstance(setup_dockerfile_commands, list):
            raise ValidationError(
                "setup_dockerfile_commands must be a list",
                field="setup_dockerfile_commands",
                value=type(setup_dockerfile_commands).__name__,
            )

        @_dockerfile_function_rep("from_registry")
        def build_dockerfile():
            commands = _Image._registry_setup_commands(
                tag, setup_dockerfile_commands, add_python
            )
            return DockerfileSpec(commands=commands, context_files={})

        return _Image._from_args(
            dockerfile_function=build_dockerfile,
            force_build=force_build,
        )

    def add_local_file(
        self, local_path: Union[str, Path], remote_path: str, *, copy: bool = True
    ) -> "_Image":
        """Adds a local file to the image at `remote_path` within the container"""
        # Validate that copy=True (Targon doesn't support lazy mounting yet)
        if not copy:
            raise ValidationError(
                "Targon SDK currently only supports copy=True for add_local_file(). "
                "Files are always copied into the image layer at build time."
            )

        # Validate remote_path is absolute
        if not PurePosixPath(remote_path).is_absolute():
            raise ValidationError(
                "image.add_local_file() currently only supports absolute remote_path values"
            )

        # Convert local_path to Path object
        local_file = Path(local_path).expanduser().resolve()

        # Validate local file exists
        if not local_file.exists():
            raise ValidationError(f"Local file not found: {local_path}")

        if not local_file.is_file():
            raise ValidationError(f"Path is not a file: {local_path}")

        # If remote_path ends with "/", append the filename
        if remote_path.endswith("/"):
            remote_path = remote_path + local_file.name

        # Create a unique context file name to avoid collisions
        context_filename = f".local_file_{local_file.name}"

        @_dockerfile_function_rep("add_local_file")
        def build_dockerfile():
            commands = [
                f"COPY {context_filename} {remote_path}",
            ]
            context_files = {context_filename: str(local_file)}
            return DockerfileSpec(commands=commands, context_files=context_files)

        return _Image._from_args(
            base_images={"base": self},
            dockerfile_function=build_dockerfile,
        )

    def add_local_dir(
        self,
        local_path: Union[str, Path],
        remote_path: str,
        *,
        copy: bool = True,
        ignore: Union[Sequence[str], Callable[[Path], bool]] = ["__pycache__"],
    ) -> "_Image":
        """Adds a local directory's content to the image at `remote_path` within the container."""
        # Validate that copy=True (Targon doesn't support lazy mounting yet)
        if not copy:
            raise ValidationError(
                "Targon SDK currently only supports copy=True for add_local_dir(). "
                "Files are always copied into the image layer at build time."
            )

        # Validate remote_path is absolute
        if not PurePosixPath(remote_path).is_absolute():
            raise ValidationError(
                "image.add_local_dir() currently only supports absolute remote_path values"
            )

        # Convert local_path to Path object
        local_dir = Path(local_path).expanduser().resolve()

        # Validate local directory exists
        if not local_dir.exists():
            raise ValidationError(f"Local directory not found: {local_path}")

        if not local_dir.is_dir():
            raise ValidationError(f"Path is not a directory: {local_path}")

        # Create ignore function
        ignore_fn = _create_ignore_function(ignore)

        # Collect all files from directory
        collected_files = _collect_directory_files(local_dir, ignore_fn)

        if not collected_files:
            raise ValidationError(
                f"No files found in directory (or all ignored): {local_path}"
            )

        # Create context files mapping with unique names
        context_files = {}
        copy_commands = []
        parent_dirs = set()

        for local_file_path, rel_path in collected_files:
            # Create a unique context filename to avoid collisions
            # Use a safe version of the relative path
            safe_name = str(rel_path).replace("/", "_").replace("\\", "_")
            context_filename = f".local_dir_{safe_name}"

            # Map context filename to actual file path
            context_files[context_filename] = str(local_file_path)

            # Create remote path preserving directory structure
            remote_file_path = PurePosixPath(remote_path) / rel_path.as_posix()

            # Collect parent directory
            parent_dirs.add(str(remote_file_path.parent))

            # Store copy command (we'll add mkdir commands before these)
            copy_commands.append(f"COPY {context_filename} {remote_file_path}")

        # Create all parent directories at once before any COPY commands
        if parent_dirs:
            # Sort for consistent ordering and remove duplicates
            sorted_dirs = sorted(parent_dirs)
            mkdir_cmd = f"RUN mkdir -p {' '.join(sorted_dirs)}"
            copy_commands.insert(0, mkdir_cmd)

        @_dockerfile_function_rep("add_local_dir")
        def build_dockerfile():
            return DockerfileSpec(commands=copy_commands, context_files=context_files)

        return _Image._from_args(
            base_images={"base": self},
            dockerfile_function=build_dockerfile,
        )

    def _inject_runtime(
        self, app_file: Union[str, Path], app_module_name: str = "app"
    ) -> "_Image":
        """
        Internal method to inject runtime into an image.
        Called automatically during image hydration when used with an App.
        """
        # Get path to targon runtime and proto files
        targon_root = Path(__file__).parent.parent  # src/targon/
        runtime_dir = targon_root / "runtime"
        proto_dir = targon_root / "proto"

        # Validate files exist
        runtime_file = runtime_dir / "universal_runtime.py"
        asgi_file = runtime_dir / "asgi.py"
        runtime_init_file = runtime_dir / "__init__.py"

        if not runtime_file.exists():
            raise ValidationError(f"Runtime file not found: {runtime_file}")
        if not asgi_file.exists():
            raise ValidationError(f"ASGI file not found: {asgi_file}")
        if not proto_dir.exists():
            raise ValidationError(f"Proto directory not found: {proto_dir}")

        app_file_path = Path(app_file).resolve()
        if not app_file_path.exists():
            raise ValidationError(f"App file not found: {app_file}")

        @_dockerfile_function_rep("universal_runtime")
        def build_dockerfile():
            commands = [
                # Install all dependencies in one layer for efficiency
                "RUN pip install --no-cache-dir \\\n"
                "    'grpcio>=1.60.0' \\\n"
                "    'grpcio-health-checking>=1.60.0' \\\n"
                "    'fastapi>=0.100.0' \\\n"
                "    'uvicorn[standard]>=0.23.0' \\\n"
                "    'cloudpickle>=2.0.0' \\\n"
                "    'protobuf>=4.25.0' \\\n"
                "    'a2wsgi>=1.10.0' \\\n"
                "    'httpx>=0.24.0'",
                # Download grpc_health_probe for Knative
                "RUN apt-get update && apt-get install -y wget && \\",
                "    wget -qO /bin/grpc_health_probe https://github.com/grpc-ecosystem/grpc-health-probe/releases/download/v0.4.19/grpc_health_probe-linux-amd64 && \\",
                "    chmod +x /bin/grpc_health_probe && \\",
                "    apt-get clean && rm -rf /var/lib/apt/lists/*",
                # Create app directory structure
                "RUN mkdir -p /app/targon/proto /app/targon/runtime",
                # Copy runtime files
                "COPY universal_runtime.py /app/targon/runtime/universal_runtime.py",
                "COPY asgi.py /app/targon/runtime/asgi.py",
                "COPY runtime_init.py /app/targon/runtime/__init__.py",
                # Copy proto files
                "COPY function_execution_pb2.py /app/targon/proto/function_execution_pb2.py",
                "COPY function_execution_pb2_grpc.py /app/targon/proto/function_execution_pb2_grpc.py",
                "COPY proto_init.py /app/targon/proto/__init__.py",
                # Copy targon package files
                "COPY targon_init.py /app/targon/__init__.py",
                "COPY version.py /app/targon/version.py",
                # Copy user's app file
                f"COPY {app_file_path.name} /app/{app_module_name}.py",
                # Set up environment
                "WORKDIR /app",
                "ENV PYTHONPATH=/app",
                f"ENV APP_MODULE={app_module_name}",
                # Expose both gRPC and web ports
                "EXPOSE 50051",
                "EXPOSE 8080",
                # Set entrypoint - TARGET_FUNCTION will be set by Knative at deployment time
                'ENTRYPOINT ["python", "-m", "targon.runtime.universal_runtime"]',
            ]

            # Map context files
            context_files = {
                # Runtime files
                "universal_runtime.py": str(runtime_file),
                "asgi.py": str(asgi_file),
                "runtime_init.py": str(runtime_init_file),
                # Proto files
                "function_execution_pb2.py": str(
                    proto_dir / "function_execution_pb2.py"
                ),
                "function_execution_pb2_grpc.py": str(
                    proto_dir / "function_execution_pb2_grpc.py"
                ),
                "proto_init.py": str(proto_dir / "__init__.py"),
                # Targon package files
                "targon_init.py": str(targon_root / "runtime" / "runtime_init.py"),
                "version.py": str(targon_root / "version.py"),
                # User app
                app_file_path.name: str(app_file_path),
            }

            return DockerfileSpec(commands=commands, context_files=context_files)

        img = _Image._from_args(
            base_images={"base": self},
            dockerfile_function=build_dockerfile,
        )
        img._runtime_injected = True
        return img

    def with_runtime(
        self, app_file: Union[str, Path], app_module_name: str = "app"
    ) -> "_Image":
        return self._inject_runtime(app_file, app_module_name)


Image = _Image
